import itertools
import logging
import random
from itertools import product

import numpy as np
import pandas as pd
import pytest

import darts.metrics as metrics
from darts import TimeSeries
from darts.datasets import AirPassengersDataset, MonthlyMilkDataset
from darts.logging import get_logger
from darts.models import (
    ARIMA,
    FFT,
    ExponentialSmoothing,
    LinearRegressionModel,
    NaiveDrift,
    NaiveSeasonal,
    RandomForest,
    Theta,
)
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils.timeseries_generation import constant_timeseries as ct
from darts.utils.timeseries_generation import gaussian_timeseries as gt
from darts.utils.timeseries_generation import linear_timeseries as lt
from darts.utils.timeseries_generation import random_walk_timeseries as rt
from darts.utils.timeseries_generation import sine_timeseries as st
from darts.utils.utils import generate_index

logger = get_logger(__name__)


if TORCH_AVAILABLE:
    from darts.models import BlockRNNModel, TCNModel


def get_dummy_series(
    ts_length: int, lt_end_value: int = 10, st_value_offset: int = 10
) -> TimeSeries:
    return (
        lt(length=ts_length, end_value=lt_end_value)
        + st(length=ts_length, value_y_offset=st_value_offset)
        + rt(length=ts_length)
    )


def compare_best_against_random(model_class, params, series, stride=1):
    # instantiate best model in expanding window mode
    np.random.seed(1)
    best_model_1, _, _ = model_class.gridsearch(
        params,
        series,
        forecast_horizon=10,
        stride=stride,
        metric=metrics.mape,
        start=series.time_index[-21],
    )

    # instantiate best model in split mode
    train, val = series.split_before(series.time_index[-10])
    best_model_2, _, _ = model_class.gridsearch(
        params, train, val_series=val, metric=metrics.mape
    )

    # instantiate model with random parameters from 'params'
    random.seed(1)
    random_param_choice = {}
    for key in params.keys():
        random_param_choice[key] = random.choice(params[key])
    random_model = model_class(**random_param_choice)

    # perform backtest forecasting on both models
    best_score_1 = best_model_1.backtest(
        series, start=series.time_index[-21], forecast_horizon=10
    )
    random_score_1 = random_model.backtest(
        series, start=series.time_index[-21], forecast_horizon=10
    )

    # perform train/val evaluation on both models
    best_model_2.fit(train)
    best_score_2 = metrics.mape(best_model_2.predict(len(val)), series)
    random_model = model_class(**random_param_choice)
    random_model.fit(train)
    random_score_2 = metrics.mape(random_model.predict(len(val)), series)

    # check whether best models are at least as good as random models
    expanding_window_ok = best_score_1 <= random_score_1
    split_ok = best_score_2 <= random_score_2

    return expanding_window_ok and split_ok


class TestBacktesting:
    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [False, True],
            [[metrics.mape], [metrics.mape, metrics.mape]],
        ),
    )
    def test_output_single_series_hfc_lpo_true(self, config):
        """Tests backtest based on historical forecasts generated on a single `series` (or list of one `series`)
        with last_points_only=True"""
        is_univariate, series_as_list, metric = config
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        y = y if not series_as_list else [y]
        hfc = hfc if not series_as_list else [hfc]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=False,
            )
        if series_as_list:
            error_msg = "Expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]`"
        else:
            error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        if series_as_list:
            with pytest.raises(ValueError) as err:
                _ = model.backtest(
                    series=y,
                    historical_forecasts=hfc + y,
                    reduction=None,
                    metric=metric,
                    last_points_only=True,
                )
            error_msg = f"expected `historical_forecasts` of type `Sequence[TimeSeries]` with length n={len(y)}."
            assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=None,
            metric=metric,
            last_points_only=True,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        if not is_multi_metric:
            # inner type expected: 1 float
            assert isinstance(bt, float) and bt == 100.0
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt, np.ndarray)
            np.testing.assert_array_almost_equal(bt, np.array([100.0, 100.0]))

        # with reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=np.mean,
            metric=metric,
            last_points_only=True,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        if not is_multi_metric:
            # inner type expected: 1 float
            assert isinstance(bt, float) and bt == 100.0
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt, np.ndarray)
            np.testing.assert_array_almost_equal(bt, np.array([100.0, 100.0]))

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [False, True],
            [[metrics.mape], [metrics.mape, metrics.mape]],
            [1, 2],
        ),
    )
    def test_output_single_series_hfc_lpo_false(self, config):
        """Tests backtest based on historical forecasts generated on a single `series` (or list of one `series`)
        with last_points_only=False"""
        is_univariate, series_as_list, metric, n_forecasts = config
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [y, hfc]
        hfc = hfc[:n_forecasts]

        y = y if not series_as_list else [y]
        hfc = hfc if not series_as_list else [hfc]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=True``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        if series_as_list:
            error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        else:
            error_msg = "Expected `historical_forecasts` of type `TimeSeries`"
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        if series_as_list:
            with pytest.raises(ValueError) as err:
                _ = model.backtest(
                    series=y,
                    historical_forecasts=hfc + [y],
                    reduction=None,
                    metric=metric,
                    last_points_only=False,
                )
            error_msg = (
                f"expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]`"
                f" with length n={len(y)}."
            )
            assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=None,
            metric=metric,
            last_points_only=False,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        assert isinstance(bt, np.ndarray)
        if not is_multi_metric:
            # inner shape expected: (n hist forecasts = 2,)
            np.testing.assert_array_almost_equal(
                bt, np.array([0.0, 100.0])[:n_forecasts]
            )
        else:
            # inner shape expected: (n hist forecasts = 2, n metrics = 2)
            np.testing.assert_array_almost_equal(
                bt, np.array([[0.0, 0.0], [100.0, 100.0]])[:n_forecasts]
            )

        # with reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=np.mean,
            metric=metric,
            last_points_only=False,
        )
        bt = bt if series_as_list else [bt]
        assert isinstance(bt, list) and len(bt) == 1
        bt = bt[0]
        score_exp = 0.0 if n_forecasts == 1 else 50.0
        if not is_multi_metric:
            # inner shape expected: 1 float
            assert isinstance(bt, float) and bt == score_exp
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt, np.ndarray)
            np.testing.assert_array_almost_equal(bt, np.array([score_exp, score_exp]))

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [[metrics.mape], [metrics.mape, metrics.mape]],
        ),
    )
    def test_output_multi_series_hfc_lpo_true(self, config):
        """Tests backtest based on historical forecasts generated on multiple `series` with last_points_only=True"""
        is_univariate, metric = config
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [y, hfc]
        y = [y, y]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=False,
            )
        error_msg = (
            "Expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]`"
        )
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc + [y[0]],
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        error_msg = f"expected `historical_forecasts` of type `Sequence[TimeSeries]` with length n={len(y)}."
        assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=None,
            last_points_only=True,
            metric=metric,
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # per series, inner type expected: 1 float
            assert bt == [0.0, 100.0]
        else:
            # per series, inner shape expected: (n metrics = 2,)
            assert all(isinstance(bt_, np.ndarray) for bt_ in bt)
            np.testing.assert_array_almost_equal(bt[0], np.array([0.0, 0.0]))
            np.testing.assert_array_almost_equal(bt[1], np.array([100.0, 100.0]))

        # with reduction
        bt = model.backtest(
            series=y,
            historical_forecasts=hfc,
            reduction=np.mean,
            last_points_only=True,
            metric=metric,
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # per series, inner type expected: 1 float
            assert bt == [0.0, 100.0]
        else:
            # per series, inner shape expected: (n metrics = 2,)
            assert all(isinstance(bt_, np.ndarray) for bt_ in bt)
            np.testing.assert_array_almost_equal(bt[0], np.array([0.0, 0.0]))
            np.testing.assert_array_almost_equal(bt[1], np.array([100.0, 100.0]))

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [[metrics.mape], [metrics.mape, metrics.mape]],
        ),
    )
    def test_output_multi_series_hfc_lpo_false(self, config):
        """Tests backtest based on historical forecasts generated on multiple `series` with
        last_points_only=False.
        """
        is_univariate, metric = config
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [[y], [hfc]]
        y = [y, y]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        assert str(err.value).startswith(error_msg)

        # number of forecasts do not match number of `series`
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc + [[y[0]]],
                reduction=None,
                metric=metric,
                last_points_only=False,
            )
        error_msg = f"expected `historical_forecasts` of type `Sequence[Sequence[TimeSeries]]` with length n={len(y)}."
        assert str(err.value).endswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=None, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        assert isinstance(bt[0], np.ndarray)
        assert isinstance(bt[1], np.ndarray)
        if not is_multi_metric:
            # inner shape expected: (n hist forecasts = 1,)
            np.testing.assert_array_almost_equal(bt[0], np.array([0.0]))
            np.testing.assert_array_almost_equal(bt[1], np.array([100.0]))
        else:
            # inner shape expected: (n metrics = 2, n hist forecasts = 1)
            np.testing.assert_array_almost_equal(bt[0], np.array([[0.0, 0.0]]))
            np.testing.assert_array_almost_equal(bt[1], np.array([[100.0, 100.0]]))

        # with reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=np.mean, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # inner type expected: 1 float
            assert bt == [0.0, 100.0]
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt[0], np.ndarray)
            np.testing.assert_array_almost_equal(bt[0], np.array([0.0, 0.0]))
            assert isinstance(bt[1], np.ndarray)
            np.testing.assert_array_almost_equal(bt[1], np.array([100.0, 100.0]))

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],
            [[metrics.mape], [metrics.mape, metrics.mape]],
        ),
    )
    def test_output_multi_series_hfc_lpo_false_different_n_fcs(self, config):
        """Tests backtest based on historical forecasts generated on multiple `series` with
        last_points_only=False, and the historical forecasts have different lengths
        """
        is_univariate, metric = config
        is_multi_metric = len(metric) > 1
        y = ct(value=1.0, length=10)
        hfc = ct(value=2.0, length=10)
        if not is_univariate:
            y = y.stack(y + 1.0)
            hfc = hfc.stack(hfc + 2.0)
        hfc = [[y], [hfc, hfc]]
        y = [y, y]

        model = NaiveDrift()

        # check that input does not work with `last_points_only=False``
        with pytest.raises(ValueError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                reduction=None,
                metric=metric,
                last_points_only=True,
            )
        error_msg = "Expected `historical_forecasts` of type `Sequence[TimeSeries]`"
        assert str(err.value).startswith(error_msg)

        # no reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=None, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        assert isinstance(bt[0], np.ndarray)
        assert isinstance(bt[1], np.ndarray)
        if not is_multi_metric:
            # inner shape expected: (n hist forecasts = 1,)
            np.testing.assert_array_almost_equal(bt[0], np.array([0.0]))
            # inner shape expected: (n hist forecasts = 2,)
            np.testing.assert_array_almost_equal(bt[1], np.array([100.0, 100.0]))
        else:
            # inner shape expected: (n metrics = 2, n hist forecasts = 1)
            np.testing.assert_array_almost_equal(bt[0], np.array([[0.0, 0.0]]))
            # inner shape expected: (n metrics = 2, n hist forecasts = 2)
            np.testing.assert_array_almost_equal(
                bt[1], np.array([[100.0, 100.0], [100.0, 100.0]])
            )

        # with reduction
        bt = model.backtest(
            series=y, historical_forecasts=hfc, reduction=np.mean, metric=metric
        )
        assert isinstance(bt, list) and len(bt) == 2
        if not is_multi_metric:
            # inner type expected: 1 float
            assert bt == [0.0, 100.0]
        else:
            # inner shape expected: (n metrics = 2,)
            assert isinstance(bt[0], np.ndarray)
            np.testing.assert_array_almost_equal(bt[0], np.array([0.0, 0.0]))
            assert isinstance(bt[1], np.ndarray)

    def test_backtest_forecasting(self):
        linear_series = lt(length=50)
        linear_series_int = TimeSeries.from_values(linear_series.values())
        linear_series_multi = linear_series.stack(linear_series)

        # univariate model + univariate series
        score = NaiveDrift().backtest(
            linear_series,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score == 1.0

        # univariate model + univariate series + historical_forecasts precalculated
        forecasts = NaiveDrift().historical_forecasts(
            linear_series,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            last_points_only=False,
        )
        precalculated_forecasts_score = NaiveDrift().backtest(
            linear_series,
            historical_forecasts=forecasts,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score == precalculated_forecasts_score

        # very large train length should not affect the backtest
        score = NaiveDrift().backtest(
            linear_series,
            train_length=10000,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score == 1.0

        # using several metric function should not affect the backtest
        score = NaiveDrift().backtest(
            linear_series,
            train_length=10000,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=[metrics.r2_score, metrics.mape],
        )
        np.testing.assert_almost_equal(score, np.array([1.0, 0.0]))

        # window of size 2 is too small for naive drift
        with pytest.raises(ValueError):
            NaiveDrift().backtest(
                linear_series,
                train_length=2,
                start=pd.Timestamp("20000201"),
                forecast_horizon=3,
                metric=metrics.r2_score,
            )

        # test that it also works for time series that are not Datetime-indexed
        score = NaiveDrift().backtest(
            linear_series_int, start=0.7, forecast_horizon=3, metric=metrics.r2_score
        )
        assert score == 1.0

        with pytest.raises(ValueError):
            NaiveDrift().backtest(
                linear_series,
                start=pd.Timestamp("20000218"),
                forecast_horizon=3,
                overlap_end=False,
            )
        NaiveDrift().backtest(
            linear_series, start=pd.Timestamp("20000217"), forecast_horizon=3
        )
        NaiveDrift().backtest(
            linear_series,
            start=pd.Timestamp("20000218"),
            forecast_horizon=3,
            overlap_end=True,
        )

        # Using forecast_horizon default value
        NaiveDrift().backtest(linear_series, start=pd.Timestamp("20000216"))
        NaiveDrift().backtest(
            linear_series, start=pd.Timestamp("20000217"), overlap_end=True
        )

        # Using an int or float value for start
        NaiveDrift().backtest(linear_series, start=30)
        NaiveDrift().backtest(linear_series, start=0.7, overlap_end=True)

        # Set custom train window length
        NaiveDrift().backtest(linear_series, train_length=10, start=30)

        # Using invalid start and/or forecast_horizon values
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=0.7, forecast_horizon=-1)
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=-0.7, forecast_horizon=1)

        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=100)
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, start=1.2)
        with pytest.raises(TypeError):
            NaiveDrift().backtest(linear_series, start="wrong type")
        with pytest.raises(ValueError):
            NaiveDrift().backtest(linear_series, train_length=0, start=0.5)
        with pytest.raises(TypeError):
            NaiveDrift().backtest(linear_series, train_length=1.2, start=0.5)
        with pytest.raises(TypeError):
            NaiveDrift().backtest(linear_series, train_length="wrong type", start=0.5)

        with pytest.raises(ValueError):
            NaiveDrift().backtest(
                linear_series, start=49, forecast_horizon=2, overlap_end=False
            )

        # univariate model + multivariate series
        with pytest.raises(ValueError):
            FFT().backtest(
                linear_series_multi, start=pd.Timestamp("20000201"), forecast_horizon=3
            )

        # multivariate model + univariate series
        if TORCH_AVAILABLE:
            tcn_model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=1,
                batch_size=1,
                n_epochs=1,
                **tfm_kwargs,
            )
            # cannot perform historical forecasts with `retrain=False` and untrained model
            with pytest.raises(ValueError):
                _ = tcn_model.historical_forecasts(
                    linear_series,
                    start=pd.Timestamp("20000125"),
                    forecast_horizon=3,
                    verbose=False,
                    last_points_only=True,
                    retrain=False,
                )

            pred = tcn_model.historical_forecasts(
                linear_series,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
                last_points_only=True,
            )
            assert pred.width == 1
            assert pred.end_time() == linear_series.end_time()

            # multivariate model + multivariate series
            # historical forecasts doesn't overwrite model object -> we can use different input dimensions
            tcn_model.backtest(
                linear_series_multi,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
            )

            # univariate model
            tcn_model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=1,
                batch_size=1,
                n_epochs=1,
                **tfm_kwargs,
            )
            tcn_model.fit(linear_series, verbose=False)
            # univariate fitted model + multivariate series
            with pytest.raises(ValueError):
                tcn_model.backtest(
                    linear_series_multi,
                    start=pd.Timestamp("20000125"),
                    forecast_horizon=3,
                    verbose=False,
                    retrain=False,
                )

            tcn_model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=3,
                batch_size=1,
                n_epochs=1,
                **tfm_kwargs,
            )
            pred = tcn_model.historical_forecasts(
                linear_series_multi,
                start=pd.Timestamp("20000125"),
                forecast_horizon=3,
                verbose=False,
                last_points_only=True,
            )
            assert pred.width == 2
            assert pred.end_time() == linear_series.end_time()

    def test_backtest_multiple_series(self):
        series = [AirPassengersDataset().load(), MonthlyMilkDataset().load()]
        model = NaiveSeasonal(K=1)

        error = model.backtest(
            series,
            train_length=30,
            forecast_horizon=2,
            stride=1,
            retrain=True,
            last_points_only=False,
            verbose=False,
        )

        expected = [11.63104, 6.09458]
        assert len(error) == 2
        assert round(abs(error[0] - expected[0]), 4) == 0
        assert round(abs(error[1] - expected[1]), 4) == 0

    def test_backtest_regression(self, caplog):
        np.random.seed(4)

        gaussian_series = gt(mean=2, length=50)
        sine_series = st(length=50)
        features = gaussian_series.stack(sine_series)
        features_multivariate = (
            (gaussian_series + sine_series).stack(gaussian_series).stack(sine_series)
        )
        target = sine_series

        features = features.with_columns_renamed(
            features.components, ["Value0", "Value1"]
        )

        features_multivariate = features_multivariate.with_columns_renamed(
            features_multivariate.components, ["Value0", "Value1", "Value2"]
        )

        # univariate feature test
        score = LinearRegressionModel(
            lags=None, lags_future_covariates=[0, -1]
        ).backtest(
            series=target,
            future_covariates=features,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
            last_points_only=True,
        )
        assert score > 0.9

        # univariate feature test + train length
        score = LinearRegressionModel(
            lags=None, lags_future_covariates=[0, -1]
        ).backtest(
            series=target,
            future_covariates=features,
            start=pd.Timestamp("20000201"),
            train_length=20,
            forecast_horizon=3,
            metric=metrics.r2_score,
            last_points_only=True,
        )
        assert score > 0.9

        # Using an int or float value for start
        score = RandomForest(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features,
            start=30,
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score > 0.9

        score = RandomForest(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features,
            start=0.5,
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score > 0.9

        # Using a too small start value
        warning_expected = (
            "`start` position `{0}` corresponding to time `{1}` is before the first "
            "predictable/trainable historical forecasting point for series at index: 0. Using the first historical "
            "forecasting point `2000-01-15 00:00:00` that lies a round-multiple of `stride=1` ahead of `start`. "
            "To hide these warnings, set `show_warnings=False`."
        )
        caplog.clear()
        with caplog.at_level(logging.WARNING):
            _ = RandomForest(lags=12).backtest(
                series=target, start=0, forecast_horizon=3
            )
            assert warning_expected.format(0, target.start_time()) in caplog.text
        caplog.clear()

        with caplog.at_level(logging.WARNING):
            _ = RandomForest(lags=12).backtest(
                series=target, start=0.01, forecast_horizon=3
            )
            assert warning_expected.format(0.01, target.start_time()) in caplog.text
        caplog.clear()

        # Using RandomForest's start default value
        score = RandomForest(lags=12, random_state=0).backtest(
            series=target, forecast_horizon=3, start=0.5, metric=metrics.r2_score
        )
        assert score > 0.95

        # multivariate feature test
        score = RandomForest(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        assert score > 0.94

        # multivariate feature test with train window 35
        score_35 = RandomForest(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            train_length=35,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        logger.info(
            "Score for multivariate feature test with train window 35 is: ", score_35
        )
        assert score_35 > 0.92

        # multivariate feature test with train window 45
        score_45 = RandomForest(
            lags=12, lags_future_covariates=[0, -1], random_state=0
        ).backtest(
            series=target,
            train_length=45,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
        )
        logger.info(
            "Score for multivariate feature test with train window 45 is: ", score_45
        )
        assert score_45 > 0.94
        assert score_45 > score_35

        # multivariate with stride
        score = RandomForest(
            lags=12, lags_future_covariates=[0], random_state=0
        ).backtest(
            series=target,
            future_covariates=features_multivariate,
            start=pd.Timestamp("20000201"),
            forecast_horizon=3,
            metric=metrics.r2_score,
            last_points_only=True,
            stride=3,
        )
        assert score > 0.9

    @pytest.mark.parametrize("model_cls", [Theta, ARIMA])
    def test_backtest_bad_covariates(self, model_cls):
        """Passing unsupported covariate should raise an exception"""
        series = lt(start_value=1, end_value=10, length=31, dtype="float32")
        model = model_cls()
        bt_kwargs = {"start": -1, "start_format": "position", "show_warnings": False}
        model.backtest(series=series, **bt_kwargs)

        with pytest.raises(ValueError) as msg:
            model.backtest(series=series, past_covariates=series, **bt_kwargs)
        assert str(msg.value).startswith(
            "Model cannot be fit/trained with `past_covariates`."
        )
        if not model.supports_future_covariates:
            with pytest.raises(ValueError) as msg:
                model.backtest(series=series, future_covariates=series, **bt_kwargs)
            assert str(msg.value).startswith(
                "Model cannot be fit/trained with `future_covariates`."
            )

    def test_gridsearch(self):
        np.random.seed(1)

        dummy_series = get_dummy_series(ts_length=50)
        dummy_series_int_index = TimeSeries.from_values(dummy_series.values())

        theta_params = {"theta": list(range(3, 10))}
        assert compare_best_against_random(Theta, theta_params, dummy_series)
        assert compare_best_against_random(Theta, theta_params, dummy_series_int_index)
        assert compare_best_against_random(Theta, theta_params, dummy_series, stride=2)

        fft_params = {"nr_freqs_to_keep": [10, 50, 100], "trend": [None, "poly", "exp"]}
        assert compare_best_against_random(FFT, fft_params, dummy_series)

        es_params = {"seasonal_periods": list(range(5, 10))}
        assert compare_best_against_random(
            ExponentialSmoothing, es_params, dummy_series
        )

    def test_gridsearch_metric_score(self):
        np.random.seed(1)

        model_class = Theta
        params = {"theta": list(range(3, 6))}
        dummy_series = get_dummy_series(ts_length=50)

        best_model, _, score = model_class.gridsearch(
            params,
            series=dummy_series,
            forecast_horizon=10,
            stride=1,
            start=dummy_series.time_index[-21],
        )
        recalculated_score = best_model.backtest(
            series=dummy_series,
            start=dummy_series.time_index[-21],
            forecast_horizon=10,
            stride=1,
        )

        assert score == recalculated_score, "The metric scores should match"

    def test_gridsearch_random_search(self):
        np.random.seed(1)

        dummy_series = get_dummy_series(ts_length=50)

        param_range = list(range(10, 20))
        params = {"lags": param_range}

        model = RandomForest(lags=1)
        result = model.gridsearch(
            params, dummy_series, forecast_horizon=1, n_random_samples=5
        )

        assert isinstance(result[0], RandomForest)
        assert isinstance(result[1]["lags"], int)
        assert isinstance(result[2], float)
        assert min(param_range) <= result[1]["lags"] <= max(param_range)

    def test_gridsearch_n_random_samples_bad_arguments(self):
        dummy_series = get_dummy_series(ts_length=50)

        params = {"lags": list(range(1, 11)), "past_covariates": list(range(1, 11))}

        with pytest.raises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=-5
            )
        with pytest.raises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=105
            )
        with pytest.raises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=-24.56
            )
        with pytest.raises(ValueError):
            RandomForest.gridsearch(
                params, dummy_series, forecast_horizon=1, n_random_samples=1.5
            )

    def test_gridsearch_n_random_samples(self):
        np.random.seed(1)

        params = {"lags": list(range(1, 11)), "past_covariates": list(range(1, 11))}

        params_cross_product = list(product(*params.values()))

        # Test absolute sample
        absolute_sampled_result = RandomForest._sample_params(params_cross_product, 10)
        assert len(absolute_sampled_result) == 10

        # Test percentage sample
        percentage_sampled_result = RandomForest._sample_params(
            params_cross_product, 0.37
        )
        assert len(percentage_sampled_result) == 37

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_gridsearch_n_jobs(self):
        """
        Testing that running gridsearch with multiple workers returns the same
        best_parameters as the single worker run.
        """

        rng_seed = 1

        np.random.seed(rng_seed)

        dummy_series = get_dummy_series(
            ts_length=100, lt_end_value=1, st_value_offset=0
        ).astype(np.float32)
        ts_train, ts_val = dummy_series.split_before(split_point=0.8)

        test_cases = [
            {
                "model": ARIMA,  # ExtendedForecastingModel
                "parameters": {"p": [18, 4], "q": [2, 3], "random_state": [rng_seed]},
            },
            {
                "model": BlockRNNModel,  # TorchForecastingModel
                "parameters": {
                    "input_chunk_length": [5, 10],
                    "output_chunk_length": [1, 3],
                    "n_epochs": [1, 5],
                    "random_state": [rng_seed],
                    "pl_trainer_kwargs": [tfm_kwargs["pl_trainer_kwargs"]],
                },
            },
        ]

        for test in test_cases:
            model = test["model"]
            parameters = test["parameters"]

            np.random.seed(rng_seed)
            _, best_params1, _ = model.gridsearch(
                parameters=parameters, series=ts_train, val_series=ts_val, n_jobs=1
            )

            np.random.seed(rng_seed)
            _, best_params2, _ = model.gridsearch(
                parameters=parameters, series=ts_train, val_series=ts_val, n_jobs=2
            )

            assert best_params1 == best_params2

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_gridsearch_multi(self):
        dummy_series = st(length=40, value_y_offset=10).stack(
            lt(length=40, end_value=20)
        )
        tcn_params = {
            "input_chunk_length": [12],
            "output_chunk_length": [3],
            "n_epochs": [1],
            "batch_size": [1],
            "kernel_size": [2, 3, 4],
            "pl_trainer_kwargs": [tfm_kwargs["pl_trainer_kwargs"]],
        }
        TCNModel.gridsearch(
            tcn_params, dummy_series, forecast_horizon=3, metric=metrics.mape
        )

    @pytest.mark.parametrize(
        "model_cls,parameters",
        zip([NaiveSeasonal, ARIMA], [{"K": [1, 2]}, {"p": [18, 4]}]),
    )
    def test_gridsearch_bad_covariates(self, model_cls, parameters):
        """Passing unsupported covariate should raise an exception"""
        dummy_series = get_dummy_series(
            ts_length=100, lt_end_value=1, st_value_offset=0
        ).astype(np.float32)

        ts_train, ts_val = dummy_series.split_before(split_point=0.8)

        bt_kwargs = {"start": -1, "start_format": "position", "show_warnings": False}

        model = model_cls()
        model_cls.gridsearch(
            parameters=parameters, series=ts_train, val_series=ts_val, **bt_kwargs
        )

        with pytest.raises(ValueError) as msg:
            model_cls.gridsearch(
                parameters=parameters,
                series=ts_train,
                past_covariates=dummy_series,
                val_series=ts_val,
                **bt_kwargs,
            )
        assert str(msg.value).startswith(
            "Model cannot be fit/trained with `past_covariates`."
        )
        if not model.supports_future_covariates:
            with pytest.raises(ValueError) as msg:
                model_cls.gridsearch(
                    parameters=parameters,
                    series=ts_train,
                    future_covariates=dummy_series,
                    val_series=ts_val,
                    **bt_kwargs,
                )
            assert str(msg.value).startswith(
                "Model cannot be fit/trained with `future_covariates`."
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product([True, False], [True, False]),
    )
    def test_gridsearch_sample_weight(self, config):
        """check that passing sample weights work and that it yields different results than without sample weights."""
        manual_weight, use_val_series = config
        ts = AirPassengersDataset().load()
        if manual_weight:
            sample_weight = np.linspace(0, 1, len(ts))
            sample_weight = ts.with_values(np.expand_dims(sample_weight, -1))
        else:
            sample_weight = "linear"

        parameters = {"lags": [3], "output_chunk_length": [1]}
        start_kwargs = {"start": -1, "start_format": "position"}
        gs_kwargs = {"val_series": ts} if use_val_series else {"forecast_horizon": 1}
        gs_non_weighted = LinearRegressionModel.gridsearch(
            parameters, series=ts[:-1], **start_kwargs, **gs_kwargs
        )[-1]

        gs_weighted = LinearRegressionModel.gridsearch(
            parameters,
            series=ts[:-1],
            sample_weight=sample_weight,
            **start_kwargs,
            **gs_kwargs,
        )[-1]

        # check that the predictions are different
        assert gs_weighted != gs_non_weighted

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [
                metrics.ase,
                metrics.mase,
            ],
            [1, 2],
        ),
    )
    def test_scaled_metrics(self, config):
        """Tests backtest for scaled metrics based on historical forecasts generated on a sequence
        `series` with last_points_only=False"""
        metric, m = config
        y = lt(length=20)
        hfc = lt(length=10, start=y.start_time() + 10 * y.freq)
        y = [y, y]
        hfc = [[hfc, hfc], [hfc]]

        model = NaiveDrift()
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=False,
            reduction=None,
            metric_kwargs={"m": m},
        )
        assert isinstance(bts, list) and len(bts) == 2

        bt_expected = metric(y[0], hfc[0][0], insample=y[0], m=m)
        for bt_list in bts:
            for bt in bt_list:
                np.testing.assert_array_almost_equal(bt, bt_expected)

    @pytest.mark.parametrize(
        "metric",
        [
            [metrics.mae],  # mae does not support time_reduction
            [metrics.mae, metrics.ae],  # ae supports time_reduction
        ],
    )
    def test_metric_kwargs(self, metric):
        """Tests backtest with different metric_kwargs based on historical forecasts generated on a sequence
        `series` with last_points_only=False"""
        y = lt(length=20)
        y = y.stack(y + 1.0)
        hfc = lt(length=10, start=y.start_time() + 10 * y.freq)
        hfc = hfc.stack(hfc + 1.0)
        y = [y, y]
        hfc = [[hfc, hfc], [hfc]]

        metric_kwargs = [{"component_reduction": np.median}]
        if len(metric) > 1:
            # give metric specific kwargs
            metric_kwargs.append({
                "component_reduction": np.median,
                "time_reduction": np.mean,
            })

        model = NaiveDrift()
        # backtest should fail with invalid metric kwargs (mae does not support time reduction)
        with pytest.raises(TypeError) as err:
            _ = model.backtest(
                series=y,
                historical_forecasts=hfc,
                metric=metric,
                last_points_only=False,
                reduction=None,
                metric_kwargs={
                    "component_reduction": np.median,
                    "time_reduction": np.mean,
                },
            )
        assert str(err.value).endswith("unexpected keyword argument 'time_reduction'")

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=False,
            reduction=None,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2

        # `ae` with time and component reduction is equal to `mae` with component reduction
        bt_expected = metrics.mae(y[0], hfc[0][0], component_reduction=np.median)
        for bt_list in bts:
            for bt in bt_list:
                np.testing.assert_array_almost_equal(bt, bt_expected)

        def time_reduced_metric(*args, **kwargs):
            return metrics.ae(*args, **kwargs, time_reduction=np.mean)

        # check that single kwargs can be used for all metrics if params are supported
        metric = [metric[0], time_reduced_metric]
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=False,
            reduction=None,
            metric_kwargs=metric_kwargs[0],
        )
        assert isinstance(bts, list) and len(bts) == 2

        # `ae` with time and component reduction is equal to `mae` with component reduction
        bt_expected = metrics.mae(y[0], hfc[0][0], component_reduction=np.median)
        for bt_list in bts:
            for bt in bt_list:
                np.testing.assert_array_almost_equal(bt, bt_expected)

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [
                [metrics.mae],  # mae does not support time_reduction
                [metrics.mae, metrics.ae],  # ae supports time_reduction
                [metrics.miw],  # quantile interval metric
                [metrics.miw, metrics.iw],
            ],
            [True, False],  # last_points_only
        ),
    )
    def test_metric_quantiles_lpo(self, config):
        """Tests backtest with quantile and quantile interval metrics from expected probabilistic or quantile
        historical forecasts."""
        metric, lpo = config
        is_interval_metric = metric[0].__name__ == "miw"

        q = [0.05, 0.5, 0.60, 0.95]
        q_interval = [(0.05, 0.50), (0.50, 0.60), (0.60, 0.95), (0.05, 0.60)]

        y = lt(length=20)
        y = y.stack(y + 1.0)
        hfc = TimeSeries.from_times_and_values(
            times=generate_index(start=y.start_time() + 10 * y.freq, length=10),
            values=np.random.random((10, 1, 100)),
        )
        hfc = hfc.stack(hfc + 1.0)
        y = [y, y]
        if lpo:
            hfc = [hfc, hfc]
        else:
            hfc = [[hfc, hfc], [hfc]]

        metric_kwargs = [{"component_reduction": np.median}]
        if not is_interval_metric:
            metric_kwargs[0]["q"] = q
        else:
            metric_kwargs[0]["q_interval"] = q_interval
        if len(metric) > 1:
            # give metric specific kwargs
            metric_kwargs2 = {
                "component_reduction": np.median,
                "time_reduction": np.mean,
            }
            if not is_interval_metric:
                metric_kwargs2["q"] = q
            else:
                metric_kwargs2["q_interval"] = q_interval
            metric_kwargs.append(metric_kwargs2)

        model = NaiveDrift()

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=None,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        if lpo:
            bts = [[bt] for bt in bts]
        # `ae` with time and component reduction is equal to `mae` with component reduction
        hfc_single = hfc[0][0] if not lpo else hfc[0]
        q_kwargs = {"q": q} if not is_interval_metric else {"q_interval": q_interval}
        bt_expected = metric[0](
            y[0], hfc_single, component_reduction=np.median, **q_kwargs
        )
        shape_expected = (len(q),)
        if len(metric) > 1:
            bt_expected = np.concatenate([bt_expected[:, None]] * 2, axis=1)
            shape_expected += (len(metric),)
        for bt_list in bts:
            for bt in bt_list:
                assert bt.shape == shape_expected
                np.testing.assert_array_almost_equal(bt, bt_expected)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=np.mean,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        for bt in bts:
            assert bt.shape == shape_expected
            np.testing.assert_array_almost_equal(bt, bt_expected)

        def time_reduced_metric(*args, **kwargs):
            metric_f = metrics.iw if is_interval_metric else metrics.ae
            return metric_f(*args, **kwargs, time_reduction=np.mean)

        # check that single kwargs can be used for all metrics if params are supported
        metric = [metric[0], time_reduced_metric]
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=None,
            metric_kwargs=metric_kwargs[0],
        )
        assert isinstance(bts, list) and len(bts) == 2
        if lpo:
            bts = [[bt] for bt in bts]
        # `ae` / `miw` with time and component reduction is equal to `mae` / `miw` with component reduction
        bt_expected = metric[0](
            y[0], hfc_single, component_reduction=np.median, **q_kwargs
        )
        bt_expected = np.concatenate([bt_expected[:, None]] * 2, axis=1)
        shape_expected = (len(q), len(metric))
        for bt_list in bts:
            for bt in bt_list:
                assert bt.shape == shape_expected
                np.testing.assert_array_almost_equal(bt, bt_expected)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=np.mean,
            metric_kwargs=metric_kwargs[0],
        )
        assert isinstance(bts, list) and len(bts) == 2
        for bt in bts:
            assert bt.shape == shape_expected
            np.testing.assert_array_almost_equal(bt, bt_expected)

        # without component reduction
        metric_kwargs = {"component_reduction": None}
        if not is_interval_metric:
            metric_kwargs["q"] = q
        else:
            metric_kwargs["q_interval"] = q_interval
        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=None,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        if lpo:
            bts = [[bt] for bt in bts]

        # `ae` / `iw` with time and no component reduction is equal to `mae` / `miw` without component reduction
        bt_expected = metric[0](y[0], hfc_single, **metric_kwargs)
        bt_expected = np.concatenate([bt_expected[:, None]] * 2, axis=1)
        shape_expected = (len(q) * y[0].width, len(metric))
        for bt_list in bts:
            for bt in bt_list:
                assert bt.shape == shape_expected
                np.testing.assert_array_almost_equal(bt, bt_expected)

        bts = model.backtest(
            series=y,
            historical_forecasts=hfc,
            metric=metric,
            last_points_only=lpo,
            reduction=np.mean,
            metric_kwargs=metric_kwargs,
        )
        assert isinstance(bts, list) and len(bts) == 2
        for bt in bts:
            assert bt.shape == shape_expected
            np.testing.assert_array_almost_equal(bt, bt_expected)

    @pytest.mark.parametrize(
        "config",
        product([True, False], [True, False]),
    )
    def test_backtest_sample_weight(self, config):
        """check that passing sample weights work and that it yields different results than without sample weights."""
        manual_weight, multi_series = config
        ts = AirPassengersDataset().load()
        if manual_weight:
            sample_weight = np.linspace(0, 1, len(ts))
            sample_weight = ts.with_values(np.expand_dims(sample_weight, -1))
        else:
            sample_weight = "linear"

        if multi_series:
            ts = [ts] * 2
            sample_weight = [sample_weight] * 2 if manual_weight else sample_weight

        model = LinearRegressionModel(lags=3, output_chunk_length=1)
        start_kwargs = {"start": -1, "start_format": "position"}
        bt_non_weighted = model.backtest(series=ts, **start_kwargs)

        model = LinearRegressionModel(lags=3, output_chunk_length=1)
        bt_weighted = model.backtest(
            series=ts, sample_weight=sample_weight, **start_kwargs
        )

        if not multi_series:
            bt_weighted = [bt_weighted]
            bt_non_weighted = [bt_non_weighted]

        # check that the predictions are different
        for bt_nw, bt_w in zip(bt_non_weighted, bt_weighted):
            assert bt_w != bt_nw



from darts.utils.historical_forecasts.optimized_historical_forecasts_regression import (
    _optimized_historical_forecasts_all_points,
    _optimized_historical_forecasts_last_points_only,
)
from darts.utils.historical_forecasts.utils import (
    _check_optimizable_historical_forecasts_global_models,
    _get_historical_forecast_boundaries,
    _historical_forecasts_general_checks,
    _process_historical_forecast_input,
)

__all__ = [
    "_optimized_historical_forecasts_all_points",
    "_optimized_historical_forecasts_last_points_only",
    "_check_optimizable_historical_forecasts_global_models",
    "_get_historical_forecast_boundaries",
    "_historical_forecasts_general_checks",
    "_process_historical_forecast_input",
]



from collections.abc import Sequence
from typing import Optional, Union

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not, raise_log
from darts.models.forecasting.forecasting_model import ForecastingModel
from darts.utils.statistics import stationarity_tests
from darts.utils.ts_utils import series2seq

logger = get_logger(__name__)


def process_input(
    model: ForecastingModel,
    input_type: str,
    series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    fallback_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    fallback_past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    fallback_future_covariates: Optional[
        Union[TimeSeries, Sequence[TimeSeries]]
    ] = None,
    check_component_names: bool = False,
    requires_input: bool = True,
    requires_covariates_encoding: bool = False,
    test_stationarity: bool = False,
):
    """Helper function to process and check either of the background or foreground series input to
    `_ForecastingModelExplainer`.

    If no input was provided (`series`, `past/future_covariates`), the fallback will be used for downstream tasks.
    Will raise an error if both input and fallback are not available.

    The fallback is dependent on the input type ("background" or "foreground"):
    - for background `input_type`: fallback are the series saved in fitted forecasting model
    - for foreground `input_type`: fallback are the background series from `_ForecastingModelExplainer`

    Parameters
    ----------
    model
        any `ForecastingModel`.
    input_type
        the type of input series. Either "background" or "foreground"
    series
         Optionally, one or a sequence of target `TimeSeries`.
    past_covariates
         Optionally, one or a sequence of past covariates `TimeSeries`.
    future_covariates
         Optionally, one or a sequence of future covariates `TimeSeries`.
    fallback_series
        Optionally, one or a sequence of target `TimeSeries` to fall back to in case `series` was not provided.
    fallback_past_covariates
        Optionally, one or a sequence of past covariates `TimeSeries` to fall back to in case `past_covariates` was
        not provided.
    fallback_future_covariates
        Optionally, one or a sequence of future covariates `TimeSeries` to fall back to in case `future_covariates` was
        not provided.
    check_component_names
        Whether to enforce that, in the case of multiple time series, all series of the same type (target or
        *_covariates) must have the same component names.
    requires_input
        Whether the input is required. If `True`, raises an error if no input was provided.
    requires_covariates_encoding
        Whether to apply the model's encoders to the input covariates. This should only be `True` if the
        Explainer will not call model methods `fit()` or `predict()` directly.
    test_stationarity
        Whether to raise a warning if not all components from the target `series` are stationary.
    """
    if input_type not in ["background", "foreground"]:
        raise_log(
            ValueError(
                f"Unknown `input_type='{input_type}'`. Must be one of ['background', 'foreground']."
            ),
            logger,
        )

    # if any input is given, treat it as if the input was required
    if (
        series is not None
        or past_covariates is not None
        or future_covariates is not None
    ):
        requires_input = True

    # if `series` was not passed, use the fallback input
    # - for background input type: fallback are the series saved in fitted forecasting model
    # - for foreground input type: fallback are the background series from `_ForecastingModelExplainer`
    if series is None:
        raise_if(
            (past_covariates is not None) or (future_covariates is not None),
            f"Supplied {input_type} past or future covariates but no {input_type} series. Please also provide "
            f"`{input_type}_series`.",
            logger,
        )
        if requires_input and fallback_series is None:
            error_msg = (
                "`model` was fit on multiple time series."
                if input_type == "background"
                else "no `background_series` was provided at `Explainer` creation"
            )
            raise_log(
                ValueError(f"`{input_type}_series` must be provided {error_msg}"),
                logger,
            )
        series = fallback_series
        past_covariates = fallback_past_covariates
        future_covariates = fallback_future_covariates
    # otherwise use the passed input, and generate the covariate encodings (they will be removed again later on
    # if `requires_covariates_encoding=False`)
    else:
        if model.encoders.encoding_available:
            past_covariates, future_covariates = model.generate_fit_encodings(
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
            )

    series = series2seq(series)
    past_covariates = series2seq(past_covariates)
    future_covariates = series2seq(future_covariates)

    (
        target_components,
        static_covariates_components,
        past_covariates_components,
        future_covariates_components,
    ) = get_component_names(
        series=series,
        past_covariates=past_covariates,
        future_covariates=future_covariates,
    )

    _check_valid_input(
        model,
        input_type,
        series,
        past_covariates,
        future_covariates,
        target_components,
        past_covariates_components,
        future_covariates_components,
        check_component_names=check_component_names,
        requires_input=requires_input,
        test_stationarity=test_stationarity,
    )

    # make sure to remove any encodings from covariates if downstream tasks require covariates without encodings
    if not requires_covariates_encoding and model.encoders.encoding_available:
        if past_covariates is not None and model.encoders.past_encoders:
            cov = past_covariates[0]
            encoded = model.encoders.past_components
            drop_cols = cov.components[cov.components.isin(encoded)]
            if not drop_cols.empty and len(drop_cols) == cov.n_components:
                past_covariates = None
            elif not drop_cols.empty:
                past_covariates = [
                    cov[cov.components.drop(drop_cols).tolist()]
                    for cov in past_covariates
                ]
        if future_covariates is not None and model.encoders.future_encoders:
            cov = future_covariates[0]
            encoded = model.encoders.future_components
            drop_cols = cov.components[cov.components.isin(encoded)]
            if not drop_cols.empty and len(drop_cols) == cov.n_components:
                future_covariates = None
            elif not drop_cols.empty:
                future_covariates = [
                    cov[cov.components.drop(drop_cols).tolist()]
                    for cov in future_covariates
                ]
    return (
        series,
        past_covariates,
        future_covariates,
        target_components,
        static_covariates_components,
        past_covariates_components,
        future_covariates_components,
    )


def process_horizons_and_targets(
    horizons: Optional[Union[int, Sequence[int]]] = None,
    fallback_horizon: Optional[int] = None,
    target_components: Optional[Union[str, Sequence[str]]] = None,
    fallback_target_components: Optional[Sequence[str]] = None,
    check_component_names: bool = False,
) -> tuple[Sequence[int], Sequence[str]]:
    """Processes the input horizons and target component names.

    horizons
        Optionally, an integer or sequence of integers representing the future time steps to be explained.
        `1` corresponds to the first timestamp being forecasted.
        All values must be `<=output_chunk_length` of the explained forecasting model.
    fallback_horizon
        Optionally, a horizon to fall back to in case `horizons` was not provided.
    target_components
        Optionally, a string or sequence of strings with the target components to explain.
    fallback_target_components
         Optionally, a sequence of strings to fall back to in case `target_components` was not provided.
    check_component_names
        Whether to enforce that the target components are in `fallback_target_component`.
    """

    if target_components is not None:
        if isinstance(target_components, str):
            target_components = [target_components]
        if check_component_names and fallback_target_components is not None:
            invalid_components = [
                target_name
                for target_name in target_components
                if target_name not in fallback_target_components
            ]
            raise_if(
                len(invalid_components) > 0,
                "Invalid `target_components`. The following components are not in the components of the "
                f"`background_series`: {invalid_components}. Provide some valid components from: "
                f"{fallback_target_components}.",
                logger,
            )
    else:
        target_components = fallback_target_components

    if horizons is not None:
        if isinstance(horizons, int):
            horizons = [horizons]

        if fallback_horizon is not None:
            raise_if(
                max(horizons) > fallback_horizon,
                "At least one of the `horizons` is larger than `output_chunk_length`.",
            )
        raise_if(min(horizons) < 1, "All `horizons` must be `>=1`.")
    else:
        horizons = range(1, fallback_horizon + 1)

    return horizons, target_components


def get_component_names(
    series: Sequence[TimeSeries],
    past_covariates: Optional[Sequence[TimeSeries]] = None,
    future_covariates: Optional[Sequence[TimeSeries]] = None,
    idx: int = 0,
) -> tuple[list[str], Optional[list[str]], Optional[list[str]], Optional[list[str]]]:
    """Extract and return the components of target series, static covariate, past and future covariates series.

    Parameters
    ----------
    model
        any `ForecastingModel`.
    series
         A sequence of target `TimeSeries`.
    past_covariates
         Optionally, a sequence of past covariates `TimeSeries`.
    future_covariates
         Optionally, a sequence of future covariates `TimeSeries`.
    idx
        the index of the input sequences to extract the components from.
    """
    target_components = series[idx].components.tolist()

    # covariates
    static_covariates = series[idx].static_covariates
    sc_components = (
        static_covariates.columns.tolist() if static_covariates is not None else []
    )
    pc_components = (
        past_covariates[idx].components.tolist() if past_covariates is not None else []
    )
    fc_components = (
        future_covariates[idx].components.tolist()
        if future_covariates is not None
        else []
    )

    # set to None if not available
    sc_components = sc_components if sc_components else None
    pc_components = pc_components if pc_components else None
    fc_components = fc_components if fc_components else None
    return target_components, sc_components, pc_components, fc_components


def _check_valid_input(
    model,
    input_type: str,
    series: Sequence[TimeSeries],
    past_covariates: Optional[Sequence[TimeSeries]],
    future_covariates: Optional[Sequence[TimeSeries]],
    target_components: Optional[list[str]],
    past_covariates_components: Optional[list[str]],
    future_covariates_components: Optional[list[str]],
    check_component_names: bool = False,
    requires_input: bool = False,
    test_stationarity: bool = False,
):
    """Checks that the input is valid"""
    if test_stationarity and series is not None:
        if not _test_stationarity(series):
            logger.warning(
                "At least one component of the target series is not stationary. "
                "Beware of wrong interpretation of the chosen explainability."
            )

    if input_type not in ["background", "foreground"]:
        raise_log(
            ValueError(
                f"Unknown `input_type='{input_type}'`. Must be one of ['background', 'foreground']."
            ),
            logger,
        )
    if past_covariates is not None:
        raise_if_not(
            len(series) == len(past_covariates),
            f"The number of {input_type} series and past covariates must be the same.",
            logger,
        )

    if future_covariates is not None:
        raise_if_not(
            len(series) == len(future_covariates),
            f"The number of {input_type} series and future covariates must be the same.",
            logger,
        )

    if requires_input:
        raise_if(
            model.uses_past_covariates and past_covariates is None,
            f"A {input_type} past covariates is not provided, but the model requires past covariates.",
            logger,
        )
        raise_if(
            model.uses_future_covariates and future_covariates is None,
            f"A {input_type} future covariates is not provided, but the model requires future covariates.",
            logger,
        )

    if not check_component_names:
        return

    # ensure we have the same names between TimeSeries (if list of). Important to ensure homogeneity
    # for explained features.
    for idx in range(len(series)):
        raise_if_not(
            all([
                series[idx].columns.to_list() == target_components,
                (
                    past_covariates[idx].columns.to_list() == past_covariates_components
                    if past_covariates is not None
                    else True
                ),
                (
                    future_covariates[idx].columns.to_list()
                    == future_covariates_components
                    if future_covariates is not None
                    else True
                ),
            ]),
            "Columns names must be identical between TimeSeries list components (multi-TimeSeries).",
        )


def _test_stationarity(series: Union[TimeSeries, Sequence[TimeSeries]]):
    return all([(stationarity_tests(bs[c]) for c in bs.components) for bs in series])



import pytest

from darts.metrics import mae
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils import timeseries_generation as tg

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )
import torch

from darts.models.forecasting.tcn_model import TCNModel


class TestTCNModel:
    def test_creation(self):
        with pytest.raises(ValueError):
            # cannot choose a kernel size larger than the input length
            TCNModel(input_chunk_length=20, output_chunk_length=1, kernel_size=100)
        TCNModel(input_chunk_length=12, output_chunk_length=1)

    def test_fit(self):
        large_ts = tg.constant_timeseries(length=100, value=1000)
        small_ts = tg.constant_timeseries(length=100, value=10)

        # Test basic fit and predict
        model = TCNModel(
            input_chunk_length=12,
            output_chunk_length=1,
            n_epochs=10,
            num_layers=1,
            **tfm_kwargs,
        )
        model.fit(large_ts[:98])
        pred = model.predict(n=2).values()[0]

        # Test whether model trained on one series is better than one trained on another
        model2 = TCNModel(
            input_chunk_length=12,
            output_chunk_length=1,
            n_epochs=10,
            num_layers=1,
            **tfm_kwargs,
        )
        model2.fit(small_ts[:98])
        pred2 = model2.predict(n=2).values()[0]
        assert abs(pred2 - 10) < abs(pred - 10)

        # test short predict
        pred3 = model2.predict(n=1)
        assert len(pred3) == 1

    def test_performance(self):
        # test TCN performance on dummy time series
        ts = tg.sine_timeseries(length=100) + tg.linear_timeseries(
            length=100, end_value=2
        )
        train, test = ts[:90], ts[90:]
        model = TCNModel(
            input_chunk_length=12,
            output_chunk_length=10,
            n_epochs=300,
            random_state=0,
            **tfm_kwargs,
        )
        model.fit(train)
        pred = model.predict(n=10)

        assert mae(pred, test) < 0.3

    @pytest.mark.slow
    def test_coverage(self):
        torch.manual_seed(0)
        input_chunk_lengths = range(20, 50)
        kernel_sizes = range(2, 5)
        dilation_bases = range(2, 5)

        for kernel_size in kernel_sizes:
            for dilation_base in dilation_bases:
                if dilation_base > kernel_size:
                    continue
                for input_chunk_length in input_chunk_lengths:
                    # create model with all weights set to one
                    model = TCNModel(
                        input_chunk_length=input_chunk_length,
                        output_chunk_length=1,
                        kernel_size=kernel_size,
                        dilation_base=dilation_base,
                        weight_norm=False,
                        n_epochs=1,
                        **tfm_kwargs,
                    )

                    # we have to fit the model on a dummy series in order to create the internal nn.Module
                    model.fit(tg.gaussian_timeseries(length=100))

                    for res_block in model.model.res_blocks:
                        res_block.conv1.weight = torch.nn.Parameter(
                            torch.ones(
                                res_block.conv1.weight.shape, dtype=torch.float64
                            )
                        )
                        res_block.conv2.weight = torch.nn.Parameter(
                            torch.ones(
                                res_block.conv2.weight.shape, dtype=torch.float64
                            )
                        )

                    model.model.eval()

                    # also disable MC Dropout:
                    model.model.set_mc_dropout(False)

                    input_tensor = torch.zeros(
                        [1, input_chunk_length, 1], dtype=torch.float64
                    )
                    zero_output = model.model.forward((input_tensor, None))[0, -1, 0]

                    # test for full coverage
                    for i in range(input_chunk_length):
                        input_tensor[0, i, 0] = 1
                        curr_output = model.model.forward((input_tensor, None))[
                            0, -1, 0
                        ]
                        assert zero_output != curr_output
                        input_tensor[0, i, 0] = 0

                    # create model with all weights set to one and one layer less than is automatically detected
                    model_2 = TCNModel(
                        input_chunk_length=input_chunk_length,
                        output_chunk_length=1,
                        kernel_size=kernel_size,
                        dilation_base=dilation_base,
                        weight_norm=False,
                        num_layers=model.model.num_layers - 1,
                        n_epochs=1,
                        **tfm_kwargs,
                    )

                    # we have to fit the model on a dummy series in order to create the internal nn.Module
                    model_2.fit(tg.gaussian_timeseries(length=100))

                    for res_block in model_2.model.res_blocks:
                        res_block.conv1.weight = torch.nn.Parameter(
                            torch.ones(
                                res_block.conv1.weight.shape, dtype=torch.float64
                            )
                        )
                        res_block.conv2.weight = torch.nn.Parameter(
                            torch.ones(
                                res_block.conv2.weight.shape, dtype=torch.float64
                            )
                        )

                    model_2.model.eval()

                    # also disable MC Dropout:
                    model_2.model.set_mc_dropout(False)

                    input_tensor = torch.zeros(
                        [1, input_chunk_length, 1], dtype=torch.float64
                    )
                    zero_output = model_2.model.forward((input_tensor, None))[0, -1, 0]

                    # test for incomplete coverage
                    uncovered_input_found = False
                    if model_2.model.num_layers == 1:
                        continue
                    for i in range(input_chunk_length):
                        input_tensor[0, i, 0] = 1
                        curr_output = model_2.model.forward((input_tensor, None))[
                            0, -1, 0
                        ]
                        if zero_output == curr_output:
                            uncovered_input_found = True
                            break
                        input_tensor[0, i, 0] = 0
                    assert uncovered_input_found

    def helper_test_pred_length(self, pytorch_model, series):
        model = pytorch_model(
            input_chunk_length=12, output_chunk_length=3, n_epochs=1, **tfm_kwargs
        )
        model.fit(series)
        pred = model.predict(7)
        assert len(pred) == 7
        pred = model.predict(2)
        assert len(pred) == 2
        assert pred.width == 1
        pred = model.predict(4)
        assert len(pred) == 4
        assert pred.width == 1

    def test_pred_length(self):
        series = tg.linear_timeseries(length=100)
        self.helper_test_pred_length(TCNModel, series)



"""
Threshold Detector
------------------

Detector that detects anomaly based on user-given threshold.
This detector compares time series values with user-given thresholds, and
identifies time points as anomalous when values are beyond the thresholds.
"""

from collections.abc import Sequence
from typing import Union

import numpy as np

from darts.ad.detectors.detectors import Detector, _BoundedDetectorMixin
from darts.logging import get_logger, raise_log
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class ThresholdDetector(Detector, _BoundedDetectorMixin):
    def __init__(
        self,
        low_threshold: Union[int, float, Sequence[float], None] = None,
        high_threshold: Union[int, float, Sequence[float], None] = None,
    ) -> None:
        """Threshold Detector

        Flags values that are either below or above the `low_threshold` and `high_threshold`,
        respectively.

        If a single value is provided for `low_threshold` or `high_threshold`, this same
        value will be used across all components of the series.

        If sequences of values are given for the parameters `low_threshold` and/or `high_threshold`,
        they must be of the same length, matching the dimensionality of the series passed
        to `detect()`, or have a length of 1. In the latter case, this single value will be used
        across all components of the series.

        If either `low_threshold` or `high_threshold` is None, the corresponding bound will not be used.
        However, at least one of the two must be set.

        Parameters
        ----------
        low_threshold
            (Sequence of) lower bounds. If a sequence, must match the dimensionality of the series this
            detector is applied to.
        high_threshold
            (Sequence of) upper bounds. If a sequence, must match the dimensionality of the series this
            detector is applied to.
        """
        super().__init__()
        low_threshold, high_threshold = self._prepare_boundaries(
            lower_bound=low_threshold,
            upper_bound=high_threshold,
            lower_bound_name="low_threshold",
            upper_bound_name="high_threshold",
        )
        self._low_threshold = low_threshold
        self._high_threshold = high_threshold

    def _detect_core(self, series: TimeSeries, name: str = "series") -> TimeSeries:
        if len(self.low_threshold) > 1 and len(self.low_threshold) != series.width:
            raise_log(
                ValueError(
                    f"The number of components for each series in `{name}` must be "
                    f"equal to the number of threshold values. Found number of "
                    f"components equal to {series.width} and expected {len(self.low_threshold)}."
                ),
                logger=logger,
            )

        # if length is 1, tile it to series width:
        low_threshold = self._expand_threshold(series[0], self.low_threshold)
        high_threshold = self._expand_threshold(series[0], self.high_threshold)

        # (time, components)
        np_series = series.values(copy=False)

        def _detect_fn(x, lo, hi):
            # x of shape (time,) for 1 component
            return (x < (-np.inf if lo is None else lo)) | (
                x > (np.inf if hi is None else hi)
            )

        detected = np.zeros_like(np_series, dtype=int)

        for component_idx in range(series.width):
            detected[:, component_idx] = _detect_fn(
                np_series[:, component_idx],
                low_threshold[component_idx],
                high_threshold[component_idx],
            )
        return series.with_values(np.expand_dims(detected, -1).astype(series.dtype))

    @property
    def low_threshold(self):
        return self._low_threshold

    @property
    def high_threshold(self):
        return self._high_threshold



"""
Forecasting Anomaly Model
-------------------------

A `ForecastingAnomalyModel` wraps around a Darts forecasting model and one or several anomaly
scorer(s) to compute anomaly scores by comparing how actuals deviate from the model's forecasts.
"""

# TODO:
#     - put start default value to its minimal value (wait for the release of historical_forecast)
import sys
from collections.abc import Sequence
from typing import Literal, Optional, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import pandas as pd

from darts.ad.anomaly_model.anomaly_model import AnomalyModel
from darts.ad.scorers.scorers import AnomalyScorer
from darts.logging import get_logger, raise_log
from darts.models.forecasting.forecasting_model import GlobalForecastingModel
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class ForecastingAnomalyModel(AnomalyModel):
    def __init__(
        self,
        model: GlobalForecastingModel,
        scorer: Union[AnomalyScorer, Sequence[AnomalyScorer]],
    ):
        """Forecasting-based Anomaly Detection Model

        The forecasting model must be a `GlobalForecastingModel` that may or may not be already fitted. The
        underlying assumption is that `model` should be able to accurately forecast the series in the absence of
        anomalies. For this reason, it is recommended to either provide a model that has already been fitted and
        evaluated to work appropriately on a series without anomalies, or to ensure that a simple call to the
        :func:`fit()` method of the model will be sufficient to train it to satisfactory performance on a series
        without anomalies. The pre-trained model will be used to generate forecasts when calling :func:`score()`.

        Calling :func:`fit()` on the anomaly model will fit the underlying forecasting model only if
        `allow_model_training` is set to `True` upon calling `fit()`.
        In addition, calling :func:`fit()` will also fit the fittable scorers, if any.

        Parameters
        ----------
        model
            An instance of a Darts forecasting model.
        scorer
            One or multiple scorer(s) that will be used to compare the actual and predicted time series in order
            to obtain an anomaly score `TimeSeries`.
            If a list of `N` scorers is given, the anomaly model will call each
            one of the scorers and output a list of `N` anomaly scores `TimeSeries`.
        """
        if not isinstance(model, GlobalForecastingModel):
            raise_log(
                ValueError("`model` must be a Darts `GlobalForecastingModel`."),
                logger=logger,
            )
        self.model = model
        super().__init__(model=model, scorer=scorer)

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        allow_model_training: bool = False,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = None,
        start_format: Literal["position", "value"] = "value",
        num_samples: int = 1,
        verbose: bool = False,
        show_warnings: bool = True,
        enable_optimization: bool = True,
        **model_fit_kwargs,
    ) -> Self:
        """Fit the underlying forecasting model (if applicable) and the fittable scorers, if any.

        Train the forecasting model (if not already fitted and `allow_model_training` is `True`) and the fittable
        scorer(s) on the given time series.

        We use the trained forecasting model to compute historical forecasts for the input `series`.
        The scorer(s) are then trained on these forecasts along with the input `series`.

        Parameters
        ----------
        series
            The (sequence of) series to train on (generally assumed to be anomaly-free).
        past_covariates
            Optionally, a (sequence of) past-observed covariate series or sequence of series. This applies only to
            models that support past covariates.
        future_covariates
            Optionally, a (sequence of) future-known covariate series or sequence of series. This applies only to
            models that support future covariates.
        allow_model_training
            Whether the forecasting model should be fitted on the given series. If `False`, the model must already be
            fitted.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        start_format
            Defines the `start` format. Only effective when `start` is an integer and `series` is indexed with a
            `pd.RangeIndex`.
            If set to 'position', `start` corresponds to the index position of the first predicted point and can range
            from `(-len(series), len(series) - 1)`.
            If set to 'value', `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: `'value'`
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Must be `1` for deterministic models.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        enable_optimization
            Whether to use the optimized version of historical_forecasts when supported and available.
            Default: ``True``.
        model_fit_kwargs
            Parameters to be passed on to the forecast model `fit()` method.

        Returns
        -------
        self
            Fitted model
        """
        return super().fit(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            allow_model_training=allow_model_training,
            forecast_horizon=forecast_horizon,
            start=start,
            start_format=start_format,
            num_samples=num_samples,
            verbose=verbose,
            show_warnings=show_warnings,
            enable_optimization=enable_optimization,
            **model_fit_kwargs,
        )

    def score(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = None,
        start_format: Literal["position", "value"] = "value",
        num_samples: int = 1,
        verbose: bool = False,
        show_warnings: bool = True,
        enable_optimization: bool = True,
        return_model_prediction: bool = False,
    ) -> Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]:
        """Compute anomaly score(s) for the given series.

        Predicts the given target time series with the forecasting model, and applies the scorer(s)
        on the prediction and the target input time series.

        Parameters
        ----------
        series
            The (sequence of) series to score on.
        past_covariates
            Optionally, a (sequence of) past-observed covariate series or sequence of series. This applies only to
            models that support past covariates.
        future_covariates
            Optionally, a (sequence of) future-known covariate series or sequence of series. This applies only to
            models that support future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        start_format
            Defines the `start` format. Only effective when `start` is an integer and `series` is indexed with a
            `pd.RangeIndex`.
            If set to 'position', `start` corresponds to the index position of the first predicted point and can range
            from `(-len(series), len(series) - 1)`.
            If set to 'value', `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: `'value'`
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Must be `1` for deterministic models.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        enable_optimization
            Whether to use the optimized version of historical_forecasts when supported and available.
            Default: ``True``.
        return_model_prediction
            Whether to return the forecasting model prediction along with the anomaly scores.

        Returns
        -------
        TimeSeries
            A single `TimeSeries` for a single `series` with a single anomaly scorers.
        Sequence[TimeSeries]
            A sequence of `TimeSeries` for:

            - a single `series` with multiple anomaly scorers.
            - a sequence of `series` with a single anomaly scorer.
        Sequence[Sequence[TimeSeries]]
            A sequence of sequences of `TimeSeries` for a sequence of `series` and multiple anomaly scorers.
            The outer sequence is over the series, and inner sequence is over the scorers.
        """
        return super().score(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=forecast_horizon,
            start=start,
            start_format=start_format,
            num_samples=num_samples,
            verbose=verbose,
            show_warnings=show_warnings,
            enable_optimization=enable_optimization,
            return_model_prediction=return_model_prediction,
        )

    def predict_series(
        self,
        series: Sequence[TimeSeries],
        past_covariates: Optional[Sequence[TimeSeries]] = None,
        future_covariates: Optional[Sequence[TimeSeries]] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = None,
        start_format: Literal["position", "value"] = "value",
        num_samples: int = 1,
        verbose: bool = False,
        show_warnings: bool = True,
        enable_optimization: bool = True,
    ) -> Sequence[TimeSeries]:
        """Computes the historical forecasts that would have been obtained by the underlying forecasting model
        on `series`.

        `retrain` is set to `False` if possible (this is not supported by all models). If set to `True`, it will always
        re-train the model on the entire available history,

        Parameters
        ----------
        series
            The sequence of series to score on.
        past_covariates
            Optionally, a sequence of past-observed covariate series or sequence of series. This applies only to
            models that support past covariates.
        future_covariates
            Optionally, a sequence of future-known covariate series or sequence of series. This applies only to
            models that support future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        start_format
            Defines the `start` format. Only effective when `start` is an integer and `series` is indexed with a
            `pd.RangeIndex`.
            If set to 'position', `start` corresponds to the index position of the first predicted point and can range
            from `(-len(series), len(series) - 1)`.
            If set to 'value', `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: `'value'`
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Must be `1` for deterministic models.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        enable_optimization
            Whether to use the optimized version of historical_forecasts when supported and available.
            Default: ``True``.

        Returns
        -------
        Sequence[TimeSeries]
            A sequence of `TimeSeries` with the historical forecasts for each series (with `last_points_only=True`).
        """
        if not self.model._fit_called:
            raise_log(
                ValueError(
                    f"Forecasting `model` {self.model} has not been trained yet. Call `fit()` before."
                ),
                logger=logger,
            )
        return self.model.historical_forecasts(
            series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=forecast_horizon,
            stride=1,
            retrain=False,
            last_points_only=True,
            start=start,
            start_format=start_format,
            num_samples=num_samples,
            verbose=verbose,
            show_warnings=show_warnings,
            enable_optimization=enable_optimization,
        )

    def eval_metric(
        self,
        anomalies: Union[TimeSeries, Sequence[TimeSeries]],
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = None,
        start_format: Literal["position", "value"] = "value",
        num_samples: int = 1,
        verbose: bool = False,
        show_warnings: bool = True,
        enable_optimization: bool = True,
        metric: Literal["AUC_ROC", "AUC_PR"] = "AUC_ROC",
    ) -> Union[
        dict[str, float],
        dict[str, Sequence[float]],
        Sequence[dict[str, float]],
        Sequence[dict[str, Sequence[float]]],
    ]:
        """Compute the accuracy of the anomaly scores computed by the model.

        Predicts the `series` with the forecasting model, and applies the scorer(s) on the predicted time series
        and the given target time series. Returns the score(s) of an agnostic threshold metric, based on the anomaly
        score given by the scorer(s).

        Parameters
        ----------
        anomalies
            The (sequence of) ground truth binary anomaly series (`1` if it is an anomaly and `0` if not).
        series
            The (sequence of) series to predict anomalies on.
        past_covariates
            Optionally, a (sequence of) past-observed covariate series or sequence of series. This applies only to
            models that support past covariates.
        future_covariates
            Optionally, a (sequence of) future-known covariate series or sequence of series. This applies only to
            models that support future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        start_format
            Defines the `start` format. Only effective when `start` is an integer and `series` is indexed with a
            `pd.RangeIndex`.
            If set to 'position', `start` corresponds to the index position of the first predicted point and can range
            from `(-len(series), len(series) - 1)`.
            If set to 'value', `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: `'value'`
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Must be `1` for deterministic models.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        enable_optimization
            Whether to use the optimized version of historical_forecasts when supported and available.
            Default: ``True``.
        metric
            The name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
            Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
            Default: "AUC_ROC".

        Returns
        -------
        Dict[str, float]
            A dictionary with the resulting metrics for single univariate `series`, with keys representing the
            anomaly scorer(s), and values representing the metric values.
        Dict[str, Sequence[float]]
            Same as for `Dict[str, float]` but for multivariate `series`, and anomaly scorers that treat series
            components/columns independently (by nature of the scorer or if `component_wise=True`).
        Sequence[Dict[str, float]]
            Same as for `Dict[str, float]` but for a sequence of univariate series.
        Sequence[Dict[str, Sequence[float]]]
            Same as for `Dict[str, float]` but for a sequence of multivariate series.
        """
        return super().eval_metric(
            anomalies=anomalies,
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=forecast_horizon,
            start=start,
            start_format=start_format,
            num_samples=num_samples,
            verbose=verbose,
            show_warnings=show_warnings,
            enable_optimization=enable_optimization,
            metric=metric,
        )

    def show_anomalies(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = None,
        start_format: Literal["position", "value"] = "value",
        num_samples: int = 1,
        verbose: bool = False,
        show_warnings: bool = True,
        enable_optimization: bool = True,
        anomalies: TimeSeries = None,
        names_of_scorers: Union[str, Sequence[str]] = None,
        title: str = None,
        metric: Optional[Literal["AUC_ROC", "AUC_PR"]] = None,
        **score_kwargs,
    ):
        """Plot the results of the anomaly model.

        Computes the score on the given series input and shows the different anomaly scores with respect to time.

        The plot will be composed of the following:

        - the series itself with the output of the forecasting model.
        - the anomaly score for each scorer. The scorers with different windows will be separated.
        - the actual anomalies, if given.

        It is possible to:

        - add a title to the figure with the parameter `title`
        - give personalized names for the scorers with `names_of_scorers`
        - show the results of a metric for each anomaly score (AUC_ROC or AUC_PR), if the actual anomalies are provided.

        Parameters
        ----------
        series
            The series to visualize anomalies from.
        past_covariates
            Optionally, a past-observed covariate series or sequence of series. This applies only to
            models that support past covariates.
        future_covariates
            Optionally, a future-known covariate series or sequence of series. This applies only to models that support
            future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        start_format
            Defines the `start` format. Only effective when `start` is an integer and `series` is indexed with a
            `pd.RangeIndex`.
            If set to 'position', `start` corresponds to the index position of the first predicted point and can range
            from `(-len(series), len(series) - 1)`.
            If set to 'value', `start` corresponds to the index value/label of the first predicted point. Will raise
            an error if the value is not in `series`' index. Default: `'value'`
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Must be `1` for deterministic models.
        verbose
            Whether to print the progress.
        show_warnings
            Whether to show warnings related to historical forecasts optimization, or parameters `start` and
            `train_length`.
        enable_optimization
            Whether to use the optimized version of historical_forecasts when supported and available.
            Default: ``True``.
        anomalies
            The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
        names_of_scorers
            Name of the scores. Must be a list of length equal to the number of scorers in the anomaly_model.
        title
            Title of the figure.
        metric
            Optionally, the name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
            Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
            Default: "AUC_ROC".
        score_kwargs
            parameters for the `score()` method.
        """
        predict_kwargs = {
            "past_covariates": past_covariates,
            "future_covariates": future_covariates,
            "forecast_horizon": forecast_horizon,
            "start": start,
            "start_format": start_format,
            "num_samples": num_samples,
            "verbose": verbose,
            "show_warnings": show_warnings,
            "enable_optimization": enable_optimization,
        }
        return super().show_anomalies(
            series=series,
            anomalies=anomalies,
            predict_kwargs=predict_kwargs,
            names_of_scorers=names_of_scorers,
            title=title,
            metric=metric,
            **score_kwargs,
        )

    def _fit_core(
        self,
        series: Sequence[TimeSeries],
        past_covariates: Optional[Sequence[TimeSeries]] = None,
        future_covariates: Optional[Sequence[TimeSeries]] = None,
        allow_model_training: bool = False,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = 0.5,
        start_format: Literal["position", "value"] = "value",
        num_samples: int = 1,
        verbose: bool = False,
        show_warnings: bool = True,
        enable_optimization: bool = True,
        **model_fit_kwargs,
    ):
        """Fit the forecasting model (if applicable) and scorers."""
        # fit forecasting model
        if allow_model_training:
            self.model._fit_wrapper(
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
                **model_fit_kwargs,
            )
        elif not self.model._fit_called:
            raise_log(
                ValueError(
                    f"With `allow_model_training=False`, the underlying model `{self.model.__class__.__name__}` "
                    f"must have already been trained. Either train it before or set `allow_model_training=True` "
                    f"(model will trained from scratch on the provided series)."
                ),
                logger=logger,
            )

        # generate the historical_forecast() prediction of the model on the train set
        if self.scorers_are_trainable:
            historical_forecasts = self.predict_series(
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
                forecast_horizon=forecast_horizon,
                start=start,
                start_format=start_format,
                num_samples=num_samples,
                verbose=verbose,
                show_warnings=show_warnings,
                enable_optimization=enable_optimization,
            )
            # fit the scorers
            self._fit_scorers(series, historical_forecasts)



import copy
import itertools
import os

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.logging import get_logger
from darts.models import (
    ExponentialSmoothing,
    LinearRegressionModel,
    NaiveDrift,
    NaiveEnsembleModel,
    NaiveSeasonal,
    RegressionEnsembleModel,
    StatsForecastAutoARIMA,
    Theta,
)
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)

if TORCH_AVAILABLE:
    from darts.models import DLinearModel, NBEATSModel, RNNModel, TCNModel
    from darts.utils.likelihood_models import QuantileRegression


def _make_ts(start_value=0, n=100):
    times = pd.date_range(start="1/1/2013", periods=n, freq="D")
    pd_series = pd.Series(range(start_value, start_value + n), index=times)
    return TimeSeries.from_series(pd_series)


class TestEnsembleModels:
    series1 = tg.sine_timeseries(value_frequency=(1 / 5), value_y_offset=10, length=50)
    series2 = tg.linear_timeseries(length=50)

    seq1 = [_make_ts(0), _make_ts(10), _make_ts(20)]
    cov1 = [_make_ts(5), _make_ts(15), _make_ts(25)]

    def test_untrained_models(self):
        model = NaiveDrift()
        _ = NaiveEnsembleModel([model])

        # trained local models should raise error
        model.fit(self.series1)
        with pytest.raises(ValueError):
            NaiveEnsembleModel([model])

        # an untrained ensemble model should also give untrained underlying models
        model_ens = NaiveEnsembleModel([NaiveDrift()])
        model_ens.fit(self.series1)
        assert model_ens.forecasting_models[0]._fit_called
        new_model = model_ens.untrained_model()
        assert not new_model.forecasting_models[0]._fit_called

    def test_trained_models(self):
        """EnsembleModels can be instantiated with pre-trained GlobalForecastingModels"""
        local_model = NaiveDrift()
        global_model = LinearRegressionModel(lags=2)
        local_model.fit(self.series1)
        global_model.fit(self.series1)

        # local and global trained
        with pytest.raises(ValueError):
            NaiveEnsembleModel([local_model, global_model])

        # local untrained, global trained
        with pytest.raises(ValueError):
            NaiveEnsembleModel([local_model.untrained_model(), global_model])

        # local trained, global untrained
        with pytest.raises(ValueError):
            NaiveEnsembleModel([local_model, global_model.untrained_model()])

        # global trained, global untrained
        with pytest.raises(ValueError):
            NaiveEnsembleModel([global_model, global_model.untrained_model()])

        # both global trained, retrain = True
        with pytest.raises(ValueError):
            # models need to be explicitly reset before retraining them
            NaiveEnsembleModel(
                [global_model, global_model], train_forecasting_models=True
            )
        model_ens_retrain = NaiveEnsembleModel(
            [global_model.untrained_model(), global_model.untrained_model()],
            train_forecasting_models=True,
        )
        with pytest.raises(ValueError):
            model_ens_retrain.predict(1, series=self.series1)
        model_ens_retrain.fit(self.series1)
        model_ens_retrain.predict(1, series=self.series1)

        # both global trained, retrain = False
        model_ens_no_retrain = NaiveEnsembleModel(
            [global_model, global_model], train_forecasting_models=False
        )
        model_ens_no_retrain.predict(1, series=self.series1)

    def test_extreme_lag_inference(self):
        ensemble = NaiveEnsembleModel([NaiveDrift()])
        assert ensemble.extreme_lags == (
            -3,
            -1,
            None,
            None,
            None,
            None,
            0,
            None,
        )  # test if default is okay

        model1 = LinearRegressionModel(
            lags=3, lags_past_covariates=[-3, -5], lags_future_covariates=[7, 8]
        )
        model2 = LinearRegressionModel(
            lags=5, lags_past_covariates=6, lags_future_covariates=[6, 9]
        )

        ensemble = NaiveEnsembleModel([
            model1,
            model2,
        ])  # test if infers extreme lags is okay
        expected = (-5, 0, -6, -1, 6, 9, 0, None)
        assert expected == ensemble.extreme_lags

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_extreme_lags_rnn(self):
        # RNNModel has the 8th element in `extreme_lags` for the `max_target_lag_train`.
        # it is given by `training_length - input_chunk_length`.
        # for the ensemble model we want the max lag of all forecasting models.
        model1 = RNNModel(input_chunk_length=14, training_length=24)
        model2 = RNNModel(input_chunk_length=12, training_length=37)

        ensemble = NaiveEnsembleModel([model1, model2])
        expected = (-14, 0, None, None, -14, 0, 0, 37 - 12)
        assert expected == ensemble.extreme_lags

    def test_input_models_local_models(self):
        with pytest.raises(ValueError):
            NaiveEnsembleModel([])
        # models are not instantiated
        with pytest.raises(ValueError):
            NaiveEnsembleModel([NaiveDrift, NaiveSeasonal, Theta, ExponentialSmoothing])
        # one model is not instantiated
        with pytest.raises(ValueError):
            NaiveEnsembleModel([
                NaiveDrift(),
                NaiveSeasonal,
                Theta(),
                ExponentialSmoothing(),
            ])
        NaiveEnsembleModel([
            NaiveDrift(),
            NaiveSeasonal(),
            Theta(),
            ExponentialSmoothing(),
        ])

    def test_call_predict_local_models(self):
        naive_ensemble = NaiveEnsembleModel([NaiveSeasonal(), Theta()])
        with pytest.raises(Exception):
            naive_ensemble.predict(5)
        naive_ensemble.fit(self.series1)
        pred1 = naive_ensemble.predict(5)
        assert self.series1.components == pred1.components

    def test_call_backtest_naive_ensemble_local_models(self):
        ensemble = NaiveEnsembleModel([NaiveSeasonal(5), Theta(2, 5)])
        ensemble.fit(self.series1)
        assert ensemble.extreme_lags == (-10, -1, None, None, None, None, 0, None)
        ensemble.backtest(self.series1)

    def test_predict_univariate_ensemble_local_models(self):
        naive = NaiveSeasonal(K=5)
        theta = Theta()
        naive_ensemble: NaiveEnsembleModel = NaiveEnsembleModel([naive, theta])
        naive_ensemble.fit(self.series1 + self.series2)
        forecast_naive_ensemble = naive_ensemble.predict(5)
        naive.fit(self.series1 + self.series2)
        theta.fit(self.series1 + self.series2)
        forecast_mean = 0.5 * naive.predict(5) + 0.5 * theta.predict(5)

        np.testing.assert_array_equal(
            forecast_naive_ensemble.values(), forecast_mean.values()
        )

    def test_predict_multivariate_ensemble_local_models(self):
        multivariate_series = self.series1.stack(self.series2)

        seasonal1 = NaiveSeasonal(K=5)
        seasonal2 = NaiveSeasonal(K=8)
        naive_ensemble: NaiveEnsembleModel = NaiveEnsembleModel([seasonal1, seasonal2])
        naive_ensemble.fit(multivariate_series)
        forecast_naive_ensemble = naive_ensemble.predict(5)
        seasonal1.fit(multivariate_series)
        seasonal2.fit(multivariate_series)
        forecast_mean = 0.5 * seasonal1.predict(5) + 0.5 * seasonal2.predict(5)

        np.testing.assert_array_equal(
            forecast_naive_ensemble.values(), forecast_mean.values()
        )
        assert all(forecast_naive_ensemble.components == multivariate_series.components)

    def test_stochastic_naive_ensemble(self):
        num_samples = 100

        # probabilistic forecasting models
        model_proba_1 = LinearRegressionModel(
            lags=1, likelihood="quantile", random_state=42
        )
        model_proba_2 = LinearRegressionModel(
            lags=2, likelihood="quantile", random_state=42
        )

        # only probabilistic forecasting models
        naive_ensemble_proba = NaiveEnsembleModel([model_proba_1, model_proba_2])
        assert naive_ensemble_proba.supports_probabilistic_prediction

        naive_ensemble_proba.fit(self.series1 + self.series2)
        # by default, only 1 sample
        pred_proba_1_sample = naive_ensemble_proba.predict(n=5)
        assert pred_proba_1_sample.n_samples == 1

        # possible to obtain probabilistic forecast by averaging samples across the models
        pred_proba_many_sample = naive_ensemble_proba.predict(
            n=5, num_samples=num_samples
        )
        assert pred_proba_many_sample.n_samples == num_samples

        # need to redefine the models to reset the random state
        model_alone_1 = LinearRegressionModel(
            lags=1, likelihood="quantile", random_state=42
        )
        model_alone_2 = LinearRegressionModel(
            lags=2, likelihood="quantile", random_state=42
        )
        model_alone_1.fit(self.series1 + self.series2)
        model_alone_2.fit(self.series1 + self.series2)
        forecast_mean = 0.5 * model_alone_1.predict(
            5, num_samples=num_samples
        ) + 0.5 * model_alone_2.predict(5, num_samples=num_samples)

        assert forecast_mean.values().shape == pred_proba_many_sample.values().shape
        assert forecast_mean.n_samples == pred_proba_many_sample.n_samples
        np.testing.assert_array_equal(
            pred_proba_many_sample.values(), forecast_mean.values()
        )

    def test_predict_likelihood_parameters_wrong_args(self):
        m_deterministic = LinearRegressionModel(lags=2, output_chunk_length=2)
        m_proba_quantile1 = LinearRegressionModel(
            lags=2,
            output_chunk_length=2,
            likelihood="quantile",
            quantiles=[0.05, 0.5, 0.95],
        )
        m_proba_quantile2 = LinearRegressionModel(
            lags=3,
            output_chunk_length=3,
            likelihood="quantile",
            quantiles=[0.05, 0.5, 0.95],
        )
        m_proba_poisson = LinearRegressionModel(
            lags=2, output_chunk_length=2, likelihood="poisson"
        )
        # one model is not probabilistic
        naive_ensemble = NaiveEnsembleModel([m_deterministic, m_proba_quantile1])
        naive_ensemble.fit(self.series1 + self.series2)
        with pytest.raises(ValueError):
            naive_ensemble.predict(n=1, predict_likelihood_parameters=True)

        # one model has a different likelihood
        naive_ensemble = NaiveEnsembleModel([
            m_proba_quantile1.untrained_model(),
            m_proba_poisson,
        ])
        naive_ensemble.fit(self.series1 + self.series2)
        with pytest.raises(ValueError):
            naive_ensemble.predict(n=1, predict_likelihood_parameters=True)

        # n > shortest output_chunk_length
        naive_ensemble = NaiveEnsembleModel([
            m_proba_quantile1.untrained_model(),
            m_proba_quantile2,
        ])
        naive_ensemble.fit(self.series1 + self.series2)
        with pytest.raises(ValueError):
            naive_ensemble.predict(n=4, predict_likelihood_parameters=True)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_predict_likelihood_parameters_univariate_naive_ensemble(self):
        m_proba_quantile1 = LinearRegressionModel(
            lags=2,
            output_chunk_length=2,
            likelihood="quantile",
            quantiles=[0.05, 0.5, 0.95],
        )
        m_proba_quantile2 = LinearRegressionModel(
            lags=3,
            output_chunk_length=2,
            likelihood="quantile",
            quantiles=[0.05, 0.5, 0.95],
        )
        m_proba_quantile3 = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=2,
            likelihood=QuantileRegression([0.05, 0.5, 0.95]),
            **tfm_kwargs,
        )

        naive_ensemble = NaiveEnsembleModel([m_proba_quantile1, m_proba_quantile2])
        naive_ensemble.fit(self.series1)
        pred_ens = naive_ensemble.predict(n=1, predict_likelihood_parameters=True)
        naive_ensemble = NaiveEnsembleModel([
            m_proba_quantile2.untrained_model(),
            m_proba_quantile3.untrained_model(),
        ])
        naive_ensemble.fit(self.series1)
        pred_mix_ens = naive_ensemble.predict(n=1, predict_likelihood_parameters=True)
        assert pred_ens.time_index == pred_mix_ens.time_index
        assert all(pred_ens.components == pred_mix_ens.components)
        assert (
            pred_ens["sine_q0.05"].values()
            < pred_ens["sine_q0.50"].values()
            < pred_ens["sine_q0.95"].values()
        )

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_predict_likelihood_parameters_multivariate_naive_ensemble(self):
        m_proba_quantile1 = LinearRegressionModel(
            lags=2,
            output_chunk_length=2,
            likelihood="quantile",
            quantiles=[0.05, 0.5, 0.95],
        )
        m_proba_quantile2 = LinearRegressionModel(
            lags=3,
            output_chunk_length=2,
            likelihood="quantile",
            quantiles=[0.05, 0.5, 0.95],
        )
        m_proba_quantile3 = DLinearModel(
            input_chunk_length=4,
            output_chunk_length=2,
            likelihood=QuantileRegression([0.05, 0.5, 0.95]),
            **tfm_kwargs,
        )

        multivariate_series = self.series1.stack(self.series2)

        naive_ensemble = NaiveEnsembleModel([m_proba_quantile1, m_proba_quantile2])
        naive_ensemble.fit(multivariate_series)
        pred_ens = naive_ensemble.predict(n=1, predict_likelihood_parameters=True)
        naive_ensemble = NaiveEnsembleModel([
            m_proba_quantile2.untrained_model(),
            m_proba_quantile3.untrained_model(),
        ])
        naive_ensemble.fit(multivariate_series)
        pred_mix_ens = naive_ensemble.predict(n=1, predict_likelihood_parameters=True)
        assert pred_ens.time_index == pred_mix_ens.time_index
        assert all(
            pred_ens.components
            == [
                "sine_q0.05",
                "sine_q0.50",
                "sine_q0.95",
                "linear_q0.05",
                "linear_q0.50",
                "linear_q0.95",
            ]
        )
        assert all(pred_ens.components == pred_mix_ens.components)
        assert (
            pred_ens["sine_q0.05"].values()
            < pred_ens["sine_q0.50"].values()
            < pred_ens["sine_q0.95"].values()
        )
        assert (
            pred_ens["linear_q0.05"].values()
            < pred_ens["linear_q0.50"].values()
            < pred_ens["linear_q0.95"].values()
        )

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_input_models_global_models(self):
        # one model is not instantiated
        with pytest.raises(ValueError):
            NaiveEnsembleModel([RNNModel(12), TCNModel(10, 2), NBEATSModel])
        NaiveEnsembleModel([RNNModel(12), TCNModel(10, 2), NBEATSModel(10, 2)])

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_call_predict_global_models_univariate_input_no_covariates(self):
        naive_ensemble = NaiveEnsembleModel([
            RNNModel(12, n_epochs=1, **tfm_kwargs),
            TCNModel(10, 2, n_epochs=1, **tfm_kwargs),
            NBEATSModel(10, 2, n_epochs=1, **tfm_kwargs),
        ])
        with pytest.raises(Exception):
            naive_ensemble.predict(5)

        naive_ensemble.fit(self.series1)
        naive_ensemble.predict(5)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_call_predict_global_models_multivariate_input_no_covariates(self):
        naive_ensemble = NaiveEnsembleModel([
            RNNModel(12, n_epochs=1, **tfm_kwargs),
            TCNModel(10, 2, n_epochs=1, **tfm_kwargs),
            NBEATSModel(10, 2, n_epochs=1, **tfm_kwargs),
        ])
        naive_ensemble.fit(self.seq1)
        naive_ensemble.predict(n=5, series=self.seq1)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_call_predict_global_models_multivariate_input_with_covariates(self):
        naive_ensemble = NaiveEnsembleModel([
            RNNModel(12, n_epochs=1, **tfm_kwargs),
            TCNModel(10, 2, n_epochs=1, **tfm_kwargs),
            NBEATSModel(10, 2, n_epochs=1, **tfm_kwargs),
        ])
        naive_ensemble.fit(self.seq1, self.cov1)
        predict_series = [s[:12] for s in self.seq1]
        predict_covariates = [c[:14] for c in self.cov1]
        naive_ensemble.predict(
            n=2, series=predict_series, past_covariates=predict_covariates
        )

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_input_models_mixed(self):
        # NaiveDrift is local, RNNModel is global
        naive_ensemble = NaiveEnsembleModel([
            NaiveDrift(),
            RNNModel(12, n_epochs=1, **tfm_kwargs),
        ])
        # ensemble is neither local, nor global
        assert not naive_ensemble.is_local_ensemble
        assert not naive_ensemble.is_global_ensemble

        # ensemble contains one local model, no support for multiple ts fit
        with pytest.raises(ValueError):
            naive_ensemble.fit([self.series1, self.series2])

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_call_predict_different_covariates_support(self):
        # AutoARIMA support future covariates only
        local_ensemble_one_covs = NaiveEnsembleModel([
            NaiveDrift(),
            StatsForecastAutoARIMA(),
        ])
        with pytest.raises(ValueError):
            local_ensemble_one_covs.fit(self.series1, past_covariates=self.series2)
        local_ensemble_one_covs.fit(self.series1, future_covariates=self.series2)

        # RNN support future covariates only
        mixed_ensemble_one_covs = NaiveEnsembleModel([
            NaiveDrift(),
            RNNModel(12, n_epochs=1, **tfm_kwargs),
        ])
        with pytest.raises(ValueError):
            mixed_ensemble_one_covs.fit(self.series1, past_covariates=self.series2)
        mixed_ensemble_one_covs.fit(self.series1, future_covariates=self.series2)

        # both models support future covariates only
        mixed_ensemble_future_covs = NaiveEnsembleModel([
            StatsForecastAutoARIMA(),
            RNNModel(12, n_epochs=1, **tfm_kwargs),
        ])
        mixed_ensemble_future_covs.fit(self.series1, future_covariates=self.series2)
        with pytest.raises(ValueError):
            mixed_ensemble_future_covs.fit(self.series1, past_covariates=self.series2)

        # RegressionModels with different covariates
        global_ensemble_both_covs = NaiveEnsembleModel([
            LinearRegressionModel(lags=1, lags_past_covariates=[-1]),
            LinearRegressionModel(lags=1, lags_future_covariates=[1]),
        ])
        # missing future covariates
        with pytest.raises(ValueError):
            global_ensemble_both_covs.fit(self.series1, past_covariates=self.series2)
        # missing past covariates
        with pytest.raises(ValueError):
            global_ensemble_both_covs.fit(self.series1, future_covariates=self.series2)
        global_ensemble_both_covs.fit(
            self.series1, past_covariates=self.series2, future_covariates=self.series2
        )

    def test_fit_multivar_ts_with_local_models(self):
        naive = NaiveEnsembleModel([
            NaiveDrift(),
            NaiveSeasonal(),
            Theta(),
            ExponentialSmoothing(),
        ])
        with pytest.raises(ValueError):
            naive.fit(self.seq1)

    def test_fit_univar_ts_with_covariates_for_local_models(self):
        naive = NaiveEnsembleModel([
            NaiveDrift(),
            NaiveSeasonal(),
            Theta(),
            ExponentialSmoothing(),
        ])
        with pytest.raises(ValueError):
            naive.fit(self.series1, self.series2)

    @pytest.mark.parametrize("model_cls", [NaiveEnsembleModel, RegressionEnsembleModel])
    def test_sample_weight_mixed_models(self, model_cls):
        """Check sample weights for ensemble models with mixed forecasting models.

        NaiveEnsembleModel
            Sample weights will only be passed to global models.
            A weighted linear model that ignores `1000` should learn that y_t = y_(t-1) + 1. When calling predict():
            - linear model should predict y_(t,lin) = 1000 + 1 = 1001
            - naive seasonal should predict y_(t,ns) = y_(t-1) = 1000

            The ensemble takes the average:
            - y_t = 0.5 * y_(t,lin) + 0.5 * y_(t,ns) = 1001 + 1000 = 1000.5

        RegressionEnsembleModel
            Sample weights will be passed to global forecasting models and regression ensemble model.
            A weighted linear model that ignores `1000` should learn that y_t = y_(t-1) + 1. When calling predict():
            - linear model should predict y_(t,lin) = y_(t-1) + 1
            - naive seasonal should predict y_(t,ns) = y_(t-1)

            The training set for regression ensemble covers the forecasts for and labels of last 5 points, where labels
            10000 and 1002 are ignored (0 weight):
            - the linear forecasting model generates forecasts: [1001, 1002, 1003, 1004, 1005]
            - the naive seasonal model generates forecasts: [1000, 1000, 1000, 1000, 1000]

            The ensemble model should then learn a perfect fit based only on the output of the linear model:
            - y_t = 1.0 * y_(t,lin) + 0.0 * y_(t,ns) = 1.0 * (y_(t-1) + 1)
            - for y_(t-1) = 1005 -> y_t = 1006
        """
        if issubclass(model_cls, NaiveEnsembleModel):
            series = TimeSeries.from_values(np.array([0.0, 1.0, 2.0, 3.0, 1000]))
            weights = TimeSeries.from_values(np.array([1.0, 1.0, 1.0, 1.0, 0.0]))
            pred_expected = np.array([[1000.5]])
            kwargs = {}
        else:
            series = TimeSeries.from_values(
                np.array([0.0, 1.0, 2.0, 3.0, 4.0, 1000, 10000, 1002, 1003, 1004, 1005])
            )
            weights = TimeSeries.from_values(
                np.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0])
            )
            pred_expected = np.array([[1006.0]])
            kwargs = {"regression_train_n_points": 5}

        model = model_cls(
            [LinearRegressionModel(lags=[-1]), NaiveSeasonal(K=1)], **kwargs
        )
        model.fit(series, sample_weight=weights)
        preds_weighted = model.predict(n=1)
        np.testing.assert_array_almost_equal(preds_weighted.values(), pred_expected)

        # make sure that without weights we get different results
        model = model_cls(
            [LinearRegressionModel(lags=[-1]), NaiveSeasonal(K=1)], **kwargs
        )
        model.fit(series)
        preds = model.predict(n=1)
        with pytest.raises(AssertionError):
            np.testing.assert_array_almost_equal(
                preds_weighted.values(), preds.values()
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product([NaiveEnsembleModel, RegressionEnsembleModel], [True, False]),
    )
    def test_sample_weight_global(self, config):
        """Check sample weights for ensemble models with global forecasting models.

        NaiveEnsembleModel
            Sample weights will only be passed to global forecasting models.
            A weighted linear model that ignores `1000` should learn that y_t = y_(t-1) + 1. When calling predict():
            - linear model should predict y_(t,lin) = 1000 + 1 = 1001

            The ensemble takes the average:
            - y_t = 0.5 * y_(t,lin) + 0.5 * y_(t,lin) = 1001 + 1001 = 1001

        RegressionEnsembleModel
            Sample weights will be passed to global forecasting models and regression ensemble model.
            A weighted linear model that ignores `1000` should learn that y_t = y_(t-1) + 1. When calling predict():
            - linear model should predict y_(t,lin) = y_(t-1) + 1

            The training set for regression ensemble covers the forecasts for and labels of last 5 points, where labels
            10000 and 1002 are ignored (0 weight):
            - the linear forecasting model generates forecasts: [1001, 1002, 1003, 1004, 1005]

            The ensemble model should then learn a perfect fit based on the output of the linear model:
            - y_t = 1.0 * y_(t,lin) + 0.0 * y_(t,ns) = 1.0 * (y_(t-1) + 1)
        """
        model_cls, single_series = config
        if issubclass(model_cls, NaiveEnsembleModel):
            series = TimeSeries.from_values(np.array([0.0, 1.0, 2.0, 3.0, 1000]))
            weights = TimeSeries.from_values(np.array([1.0, 1.0, 1.0, 1.0, 0.0]))
            pred_expected = np.array([[1001.0]])
            kwargs = {}
        else:
            series = TimeSeries.from_values(
                np.array([0.0, 1.0, 2.0, 3.0, 4.0, 1000, 10000, 1002, 1003, 1004, 1005])
            )
            weights = TimeSeries.from_values(
                np.array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0])
            )
            pred_expected = np.array([[1006.0]])
            kwargs = {"regression_train_n_points": 5}

        if not single_series:
            series = [series] * 2
            weights = [weights] * 2

        model = model_cls(
            [LinearRegressionModel(lags=[-1]), LinearRegressionModel(lags=[-1])],
            **kwargs,
        )
        model.fit(series, sample_weight=weights)
        preds_weighted = model.predict(n=1, series=series)
        if single_series:
            preds_weighted = [preds_weighted]

        for preds in preds_weighted:
            np.testing.assert_array_almost_equal(preds.values(), pred_expected)

        # make sure that without weights we get different results
        model = model_cls(
            [LinearRegressionModel(lags=[-1]), LinearRegressionModel(lags=[-1])],
            **kwargs,
        )
        model.fit(series)
        preds = model.predict(n=1, series=series)
        if single_series:
            preds = [preds]

        for pred_w, pred_nw in zip(preds_weighted, preds):
            with pytest.raises(AssertionError):
                np.testing.assert_array_almost_equal(pred_w.values(), pred_nw.values())

    @pytest.mark.parametrize("model_cls", [NaiveEnsembleModel, RegressionEnsembleModel])
    def test_invalid_sample_weight(self, model_cls):
        kwargs = {
            "forecasting_models": [
                LinearRegressionModel(lags=[-1]),
                NaiveSeasonal(K=1),
            ],
        }
        if issubclass(model_cls, RegressionEnsembleModel):
            kwargs["regression_train_n_points"] = 3

        ts = TimeSeries.from_values(np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]))
        # weights too short
        model = model_cls(**copy.deepcopy(kwargs))
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight=ts[:-1])
        assert (
            str(err.value)
            == "The `sample_weight` series must have at least the same times as the target `series`."
        )

        # same number of series
        model = model_cls(**copy.deepcopy(kwargs))
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight=[ts, ts])
        assert (
            str(err.value)
            == "The provided sequence of target `series` must have the same length as the "
            "provided sequence of `sample_weight`."
        )

        # same number of components
        model = model_cls(**copy.deepcopy(kwargs))
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight=ts.stack(ts))
        assert (
            str(err.value)
            == "The number of components in `sample_weight` must either be `1` or match the "
            "number of target series components `1`."
        )
        # with correct number it works
        model = model_cls(**copy.deepcopy(kwargs))
        model.fit(ts.stack(ts), sample_weight=ts.stack(ts))
        # or with multivar ts and single component weights (globally applied)
        model = model_cls(**copy.deepcopy(kwargs))
        model.fit(ts.stack(ts), sample_weight=ts)

        # invalid string
        model = model_cls(**copy.deepcopy(kwargs))
        with pytest.raises(ValueError) as err:
            model.fit(ts, sample_weight="invalid")
        assert str(err.value).startswith("Invalid `sample_weight` value: `'invalid'`. ")

        # but with valid string it works
        model.fit(ts, sample_weight="linear")

    def test_predict_with_target(self):
        series_long = self.series1
        series_short = series_long[:25]

        # train with a single series
        ensemble_model = self.get_global_ensemble_model()
        ensemble_model.fit(series_short, past_covariates=series_long)
        # predict after end of train series
        preds = ensemble_model.predict(n=5, past_covariates=series_long)
        assert isinstance(preds, TimeSeries)
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        assert isinstance(preds, TimeSeries)
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        assert isinstance(preds, list) and len(preds) == 2
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        assert isinstance(preds, list) and len(preds) == 1

        # train with multiple series
        ensemble_model = self.get_global_ensemble_model()
        ensemble_model.fit([series_short] * 2, past_covariates=[series_long] * 2)
        with pytest.raises(ValueError):
            # predict without passing series should raise an error
            ensemble_model.predict(n=5, past_covariates=series_long)
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        assert isinstance(preds, TimeSeries)
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        assert isinstance(preds, list) and len(preds) == 2
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        assert isinstance(preds, list) and len(preds) == 1

    @staticmethod
    def get_global_ensemble_model(output_chunk_length=5):
        lags = [-1, -2, -5]
        return NaiveEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
            ],
        )

    @pytest.mark.parametrize("model_cls", [NaiveEnsembleModel, RegressionEnsembleModel])
    def test_save_load_ensemble_models(self, tmpdir_fn, model_cls):
        # check if save and load methods work and
        # if loaded ensemble model creates same forecasts as original ensemble models
        full_model_path_str = os.getcwd()
        kwargs = {}
        expected_suffixes = [".pkl", ".pkl.RNNModel_2.pt", ".pkl.RNNModel_2.pt.ckpt"]

        if issubclass(model_cls, RegressionEnsembleModel):
            kwargs["regression_train_n_points"] = 5

        if TORCH_AVAILABLE:
            model = model_cls(
                [
                    LinearRegressionModel(lags=[-1]),
                    NaiveSeasonal(K=1),
                    RNNModel(10, n_epochs=1, **tfm_kwargs),
                ],
                **kwargs,
            )
        else:
            model = model_cls(
                [LinearRegressionModel(lags=[-1]), NaiveSeasonal(K=1)], **kwargs
            )

        model.fit(self.series1 + self.series2)
        model_prediction = model.predict(5)

        # test save
        model.save()
        model.save(os.path.join(full_model_path_str, f"{model_cls.__name__}.pkl"))

        assert os.path.exists(full_model_path_str)
        files = os.listdir(full_model_path_str)
        if TORCH_AVAILABLE:
            assert len(files) == 6
            for f in files:
                assert f.startswith(model_cls.__name__)
            suffix_counts = {
                suffix: sum(
                    1 for p in os.listdir(full_model_path_str) if p.endswith(suffix)
                )
                for suffix in expected_suffixes
            }
            assert all(count == 2 for count in suffix_counts.values())
        else:
            assert len(files) == 2
            for f in files:
                assert f.startswith(model_cls.__name__) and f.endswith(".pkl")

        # test load
        pkl_files = []
        for filename in os.listdir(full_model_path_str):
            if filename.endswith(".pkl"):
                pkl_files.append(os.path.join(full_model_path_str, filename))
        for p in pkl_files:
            loaded_model = model_cls.load(p)
            assert model_prediction == loaded_model.predict(5)



"""
Window Transformer
------------------
"""

from collections.abc import Mapping
from typing import Any, Optional, Union

from darts.dataprocessing.transformers import BaseDataTransformer
from darts.logging import get_logger
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class WindowTransformer(BaseDataTransformer):
    def __init__(
        self,
        transforms: Union[dict, list[dict]],
        treat_na: Optional[Union[str, Union[int, float]]] = None,
        forecasting_safe: Optional[bool] = True,
        keep_non_transformed: Optional[bool] = False,
        include_current: Optional[bool] = True,
        keep_names: Optional[bool] = False,
        name: str = "WindowTransformer",
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """
        A transformer that applies window transformation to a TimeSeries or a Sequence of TimeSeries. It expects a
        dictionary or a list of dictionaries specifying the window transformation(s) to be applied. All series in the
        sequence will be transformed with the same transformations.

        Parameters
        ----------
        transforms
            A dictionary or a list of dictionaries.
            Each dictionary specifies a different window transform.

            The dictionaries can contain the following keys:

            :``"function"``: Mandatory. The name of one of the pandas builtin transformation functions,
                            or a callable function that can be applied to the input series.
                            Pandas' functions can be found in the
                            `documentation <https://pandas.pydata.org/docs/reference/window.html>`_.

            :``"mode"``: Optional. The name of the pandas windowing mode on which the ``"function"`` is going to be
                        applied. The options are "rolling", "expanding" and "ewm".
                        If not provided, Darts defaults to "expanding".
                        User defined functions can use either "rolling" or "expanding" modes.
                        More information on pandas windowing operations can be found in the `documentation
                        <https://pandas.pydata.org/pandas-docs/stable/user_guide/window.html>`_.

            :``"components"``: Optional. A string or list of strings specifying the TimeSeries components on which the
                               transformation should be applied. If not specified, the transformation will be
                               applied on all components.

            :``"function_name"``: Optional. A string specifying the function name referenced as part of
                                  the transformation output name. For example, given a user-provided function
                                  transformation on rolling window size of 5 on the component "comp", the
                                  default transformation output name is "rolling_udf_5_comp" whereby "udf"
                                  refers to "user defined function". If specified, the ``"function_name"`` will
                                  replace the default name "udf". Similarly, the ``"function_name"`` will replace
                                  the name of the pandas builtin transformation function name in the output name.

            All other dictionary items provided will be treated as keyword arguments for the windowing mode
            (i.e., ``rolling/ewm/expanding``) or for the specific function
            in that mode (i.e., ``pandas.DataFrame.rolling.mean/std/max/min...`` or
            ``pandas.DataFrame.ewm.mean/std/sum``).
            This allows for more flexibility in configuring the transformation, by providing for
            example:

            * :``"window"``: Size of the moving window for the "rolling" mode.
                            If an integer, the fixed number of observations used for each window.
                            If an offset, the time period of each window with data type :class:`pandas.Timedelta`
                            representing a fixed duration.
            * :``"min_periods"``: The minimum number of observations in the window required to have a value (otherwise
                NaN). Darts reuses pandas defaults of 1 for "rolling" and "expanding" modes and of 0 for "ewm" mode.
            * :``"win_type"``: The type of weigthing to apply to the window elements.
                If provided, it should be one of `scipy.signal.windows
                <https://docs.scipy.org/doc/scipy/reference/signal.windows.html#module-scipy.signal.windows>`_.
            * :``"center"``: ``True``/``False`` to set the observation at the current timestep at the center of the
                window (when ``forecasting_safe`` is `True`, Darts enforces ``"center"`` to ``False``).
            * :``"closed"``: ``"right"``/``"left"``/``"both"``/``"neither"`` to specify whether the right,
                left or both ends of the window are included in the window, or neither of them.
                Darts defaults to ``"both"``.

            More information on the available functions and their parameters can be found in the
            `Pandas documentation <https://pandas.pydata.org/docs/reference/window.html>`_.

            For user-provided functions, extra keyword arguments in the transformation dictionary are passed to the
            user-defined function.
            By default, Darts expects user-defined functions to receive numpy arrays as input.
            This can be modified by adding item ``"raw": False`` in the transformation dictionary.
            It is expected that the function returns a single
            value for each window. Other possible configurations can be found in the
            `pandas.DataFrame.rolling().apply() documentation
            <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.rolling.html>`_
            and `pandas.DataFrame.expanding().apply() documentation
            <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.expanding.html>`_.

        treat_na
            Specifies how to treat missing values that were added by the window transformations
            at the beginning of the resulting TimeSeries. By default, Darts will leave NaNs in the resulting TimeSeries.
            This parameter can be one of the following:

            * :``"dropna"``: to truncate the TimeSeries and drop time steps containing missing values.
                If multiple columns contain different numbers of missing values, only the minimum number
                of rows is dropped. This operation might reduce the length of the resulting TimeSeries.

            * :``"bfill"`` or ``"backfill"``: to specify that NaNs should be filled with the last transformed
                and valid observation. If the original TimeSeries starts with NaNs, those are kept.
                When ``forecasting_safe`` is ``True``, this option returns an exception to avoid future observation
                contaminating the past.

            * :an integer or float: in which case NaNs will be filled with this value.
                All columns will be filled with the same provided value.

        forecasting_safe
            If True, Darts enforces that the resulting TimeSeries is safe to be used in forecasting models as target
            or as feature. The window transformation will not allow future values to be included in the computations
            at their corresponding current timestep. Default is ``True``.
            "ewm" and "expanding" modes are forecasting safe by default.
            "rolling" mode is forecasting safe if ``"center": False`` is guaranteed.

        keep_non_transformed
            ``False`` to return the transformed components only, ``True`` to return all original components along
            the transformed ones. Default is ``False``. If the series has a hierarchy, must be set to ``False``.

        include_current
            ``True`` to include the current time step in the window, ``False`` to exclude it. Default is ``True``.

        keep_names
            Whether the transformed components should keep the original component names or. Must be set to ``False``
            if `keep_non_transformed = True` or the number of transformation is greater than 1.

        name
            A specific name for the transformer.

        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`.

        verbose
            Whether to print operations progress.
        """

        # dictionary checks are implemented in TimeSeries.window_transform()

        # Define fixed params (i.e. attributes defined before calling `super().__init__`):
        self.transforms = transforms
        self.keep_non_transformed = keep_non_transformed
        self.treat_na = treat_na
        self.forecasting_safe = forecasting_safe
        self.include_current = include_current
        self.keep_names = keep_names
        super().__init__(name, n_jobs, verbose)

    @staticmethod
    def ts_transform(series: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
        return series.window_transform(**params["fixed"])
